#!/usr/bin/env python3
# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run tests in parallel."""

from __future__ import print_function

import argparse
import ast
import collections
import glob
import itertools
import json
import logging
import multiprocessing
import os
import os.path
import pipes
import platform
import random
import re
import socket
import subprocess
import sys
import tempfile
import time
import traceback
import uuid

import six
from six.moves import urllib

import python_utils.jobset as jobset
import python_utils.report_utils as report_utils
import python_utils.start_port_server as start_port_server
import python_utils.watch_dirs as watch_dirs

try:
    from python_utils.upload_test_results import upload_results_to_bq
except (ImportError):
    pass  # It's ok to not import because this is only necessary to upload results to BQ.

gcp_utils_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '../gcp/utils'))
sys.path.append(gcp_utils_dir)

_ROOT = os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]), '../..'))
os.chdir(_ROOT)

_FORCE_ENVIRON_FOR_WRAPPERS = {
    'GRPC_VERBOSITY': 'DEBUG',
}

_POLLING_STRATEGIES = {
    'linux': ['epoll1', 'poll'],
    'mac': ['poll'],
}


def platform_string():
    return jobset.platform_string()


_DEFAULT_TIMEOUT_SECONDS = 5 * 60
_PRE_BUILD_STEP_TIMEOUT_SECONDS = 10 * 60


def run_shell_command(cmd, env=None, cwd=None):
    try:
        subprocess.check_output(cmd, shell=True, env=env, cwd=cwd)
    except subprocess.CalledProcessError as e:
        logging.exception(
            "Error while running command '%s'. Exit status %d. Output:\n%s",
            e.cmd, e.returncode, e.output)
        raise


def max_parallel_tests_for_current_platform():
    # Too much test parallelization has only been seen to be a problem
    # so far on windows.
    if jobset.platform_string() == 'windows':
        return 64
    return 1024


def _print_debug_info_epilogue(dockerfile_dir=None):
    """Use to print useful info for debug/repro just before exiting."""
    print('')
    print('=== run_tests.py DEBUG INFO ===')
    print('command: \"%s\"' % ' '.join(sys.argv))
    if dockerfile_dir:
        print('dockerfile: %s' % dockerfile_dir)
    kokoro_job_name = os.getenv('KOKORO_JOB_NAME')
    if kokoro_job_name:
        print('kokoro job name: %s' % kokoro_job_name)
    print('===============================')


# SimpleConfig: just compile with CONFIG=config, and run the binary to test
class Config(object):

    def __init__(self,
                 config,
                 environ=None,
                 timeout_multiplier=1,
                 tool_prefix=[],
                 iomgr_platform='native'):
        if environ is None:
            environ = {}
        self.build_config = config
        self.environ = environ
        self.environ['CONFIG'] = config
        self.tool_prefix = tool_prefix
        self.timeout_multiplier = timeout_multiplier
        self.iomgr_platform = iomgr_platform

    def job_spec(self,
                 cmdline,
                 timeout_seconds=_DEFAULT_TIMEOUT_SECONDS,
                 shortname=None,
                 environ={},
                 cpu_cost=1.0,
                 flaky=False):
        """Construct a jobset.JobSpec for a test under this config

       Args:
         cmdline:      a list of strings specifying the command line the test
                       would like to run
    """
        actual_environ = self.environ.copy()
        for k, v in environ.items():
            actual_environ[k] = v
        if not flaky and shortname and shortname in flaky_tests:
            flaky = True
        if shortname in shortname_to_cpu:
            cpu_cost = shortname_to_cpu[shortname]
        return jobset.JobSpec(
            cmdline=self.tool_prefix + cmdline,
            shortname=shortname,
            environ=actual_environ,
            cpu_cost=cpu_cost,
            timeout_seconds=(self.timeout_multiplier *
                             timeout_seconds if timeout_seconds else None),
            flake_retries=4 if flaky or args.allow_flakes else 0,
            timeout_retries=1 if flaky or args.allow_flakes else 0)


def get_c_tests(travis, test_lang):
    out = []
    platforms_str = 'ci_platforms' if travis else 'platforms'
    with open('tools/run_tests/generated/tests.json') as f:
        js = json.load(f)
        return [
            tgt for tgt in js
            if tgt['language'] == test_lang and platform_string() in
            tgt[platforms_str] and not (travis and tgt['flaky'])
        ]


def _check_compiler(compiler, supported_compilers):
    if compiler not in supported_compilers:
        raise Exception('Compiler %s not supported (on this platform).' %
                        compiler)


def _check_arch(arch, supported_archs):
    if arch not in supported_archs:
        raise Exception('Architecture %s not supported.' % arch)


def _is_use_docker_child():
    """Returns True if running running as a --use_docker child."""
    return True if os.getenv('DOCKER_RUN_SCRIPT_COMMAND') else False


_PythonConfigVars = collections.namedtuple('_ConfigVars', [
    'shell',
    'builder',
    'builder_prefix_arguments',
    'venv_relative_python',
    'toolchain',
    'runner',
])


def _python_config_generator(name, major, minor, bits, config_vars):
    build = (config_vars.shell + config_vars.builder +
             config_vars.builder_prefix_arguments +
             [_python_pattern_function(major=major, minor=minor, bits=bits)] +
             [name] + config_vars.venv_relative_python + config_vars.toolchain)
    run = (config_vars.shell + config_vars.runner + [
        os.path.join(name, config_vars.venv_relative_python[0]),
    ])
    return PythonConfig(name, build, run)


def _pypy_config_generator(name, major, config_vars):
    return PythonConfig(
        name, config_vars.shell + config_vars.builder +
        config_vars.builder_prefix_arguments +
        [_pypy_pattern_function(major=major)] + [name] +
        config_vars.venv_relative_python + config_vars.toolchain,
        config_vars.shell + config_vars.runner +
        [os.path.join(name, config_vars.venv_relative_python[0])])


def _python_pattern_function(major, minor, bits):
    # Bit-ness is handled by the test machine's environment
    if os.name == "nt":
        if bits == "64":
            return '/c/Python{major}{minor}/python.exe'.format(major=major,
                                                               minor=minor,
                                                               bits=bits)
        else:
            return '/c/Python{major}{minor}_{bits}bits/python.exe'.format(
                major=major, minor=minor, bits=bits)
    else:
        return 'python{major}.{minor}'.format(major=major, minor=minor)


def _pypy_pattern_function(major):
    if major == '2':
        return 'pypy'
    elif major == '3':
        return 'pypy3'
    else:
        raise ValueError("Unknown PyPy major version")


class CLanguage(object):

    def __init__(self, lang_suffix, test_lang):
        self.lang_suffix = lang_suffix
        self.platform = platform_string()
        self.test_lang = test_lang

    def configure(self, config, args):
        self.config = config
        self.args = args
        if self.platform == 'windows':
            _check_compiler(self.args.compiler, [
                'default',
                'cmake',
                'cmake_ninja_vs2017',
                'cmake_vs2017',
                'cmake_vs2019',
            ])
            _check_arch(self.args.arch, ['default', 'x64', 'x86'])

            activate_vs_tools = ''
            if self.args.compiler == 'cmake_ninja_vs2017' or self.args.compiler == 'cmake' or self.args.compiler == 'default':
                # cmake + ninja build is the default because it is faster and supports boringssl assembly optimizations
                # the compiler used is exactly the same as for cmake_vs2017
                cmake_generator = 'Ninja'
                activate_vs_tools = '2017'
            elif self.args.compiler == 'cmake_vs2017':
                cmake_generator = 'Visual Studio 15 2017'
            elif self.args.compiler == 'cmake_vs2019':
                cmake_generator = 'Visual Studio 16 2019'
            else:
                print('should never reach here.')
                sys.exit(1)

            self._cmake_configure_extra_args = []
            self._cmake_generator_windows = cmake_generator
            # required to pass as cmake "-A" configuration for VS builds (but not for Ninja)
            self._cmake_architecture_windows = 'x64' if self.args.arch == 'x64' else 'Win32'
            # when builing with Ninja, the VS common tools need to be activated first
            self._activate_vs_tools_windows = activate_vs_tools
            self._vs_tools_architecture_windows = 'x64' if self.args.arch == 'x64' else 'x86'

        else:
            if self.platform == 'linux':
                # Allow all the known architectures. _check_arch_option has already checked that we're not doing
                # something illegal when not running under docker.
                _check_arch(self.args.arch, ['default', 'x64', 'x86', 'arm64'])
            else:
                _check_arch(self.args.arch, ['default'])

            self._docker_distro, self._cmake_configure_extra_args = self._compiler_options(
                self.args.use_docker, self.args.compiler)

            if self.args.arch == 'x86':
                # disable boringssl asm optimizations when on x86
                # see https://github.com/grpc/grpc/blob/b5b8578b3f8b4a9ce61ed6677e19d546e43c5c68/tools/run_tests/artifacts/artifact_targets.py#L253
                self._cmake_configure_extra_args.append('-DOPENSSL_NO_ASM=ON')

    def test_specs(self):
        out = []
        binaries = get_c_tests(self.args.travis, self.test_lang)
        for target in binaries:
            if target.get('boringssl', False):
                # cmake doesn't build boringssl tests
                continue
            auto_timeout_scaling = target.get('auto_timeout_scaling', True)
            polling_strategies = (_POLLING_STRATEGIES.get(
                self.platform, ['all']) if target.get('uses_polling', True) else
                                  ['none'])
            for polling_strategy in polling_strategies:
                env = {
                    'GRPC_DEFAULT_SSL_ROOTS_FILE_PATH':
                        _ROOT + '/src/core/tsi/test_creds/ca.pem',
                    'GRPC_POLL_STRATEGY':
                        polling_strategy,
                    'GRPC_VERBOSITY':
                        'DEBUG'
                }
                resolver = os.environ.get('GRPC_DNS_RESOLVER', None)
                if resolver:
                    env['GRPC_DNS_RESOLVER'] = resolver
                shortname_ext = '' if polling_strategy == 'all' else ' GRPC_POLL_STRATEGY=%s' % polling_strategy
                if polling_strategy in target.get('excluded_poll_engines', []):
                    continue

                timeout_scaling = 1
                if auto_timeout_scaling:
                    config = self.args.config
                    if ('asan' in config or config == 'msan' or
                            config == 'tsan' or config == 'ubsan' or
                            config == 'helgrind' or config == 'memcheck'):
                        # Scale overall test timeout if running under various sanitizers.
                        # scaling value is based on historical data analysis
                        timeout_scaling *= 3

                if self.config.build_config in target['exclude_configs']:
                    continue
                if self.args.iomgr_platform in target.get('exclude_iomgrs', []):
                    continue
                if self.platform == 'windows':
                    binary = 'cmake/build/%s/%s.exe' % (_MSBUILD_CONFIG[
                        self.config.build_config], target['name'])
                else:
                    binary = 'cmake/build/%s' % target['name']

                cpu_cost = target['cpu_cost']
                if cpu_cost == 'capacity':
                    cpu_cost = multiprocessing.cpu_count()
                if os.path.isfile(binary):
                    list_test_command = None
                    filter_test_command = None

                    # these are the flag defined by gtest and benchmark framework to list
                    # and filter test runs. We use them to split each individual test
                    # into its own JobSpec, and thus into its own process.
                    if 'benchmark' in target and target['benchmark']:
                        with open(os.devnull, 'w') as fnull:
                            tests = subprocess.check_output(
                                [binary, '--benchmark_list_tests'],
                                stderr=fnull)
                        for line in tests.decode().split('\n'):
                            test = line.strip()
                            if not test:
                                continue
                            cmdline = [binary,
                                       '--benchmark_filter=%s$' % test
                                      ] + target['args']
                            out.append(
                                self.config.job_spec(
                                    cmdline,
                                    shortname='%s %s' %
                                    (' '.join(cmdline), shortname_ext),
                                    cpu_cost=cpu_cost,
                                    timeout_seconds=target.get(
                                        'timeout_seconds',
                                        _DEFAULT_TIMEOUT_SECONDS) *
                                    timeout_scaling,
                                    environ=env))
                    elif 'gtest' in target and target['gtest']:
                        # here we parse the output of --gtest_list_tests to build up a complete
                        # list of the tests contained in a binary for each test, we then
                        # add a job to run, filtering for just that test.
                        with open(os.devnull, 'w') as fnull:
                            tests = subprocess.check_output(
                                [binary, '--gtest_list_tests'], stderr=fnull)
                        base = None
                        for line in tests.decode().split('\n'):
                            i = line.find('#')
                            if i >= 0:
                                line = line[:i]
                            if not line:
                                continue
                            if line[0] != ' ':
                                base = line.strip()
                            else:
                                assert base is not None
                                assert line[1] == ' '
                                test = base + line.strip()
                                cmdline = [binary,
                                           '--gtest_filter=%s' % test
                                          ] + target['args']
                                out.append(
                                    self.config.job_spec(
                                        cmdline,
                                        shortname='%s %s' %
                                        (' '.join(cmdline), shortname_ext),
                                        cpu_cost=cpu_cost,
                                        timeout_seconds=target.get(
                                            'timeout_seconds',
                                            _DEFAULT_TIMEOUT_SECONDS) *
                                        timeout_scaling,
                                        environ=env))
                    else:
                        cmdline = [binary] + target['args']
                        shortname = target.get(
                            'shortname',
                            ' '.join(pipes.quote(arg) for arg in cmdline))
                        shortname += shortname_ext
                        out.append(
                            self.config.job_spec(
                                cmdline,
                                shortname=shortname,
                                cpu_cost=cpu_cost,
                                flaky=target.get('flaky', False),
                                timeout_seconds=target.get(
                                    'timeout_seconds',
                                    _DEFAULT_TIMEOUT_SECONDS) * timeout_scaling,
                                environ=env))
                elif self.args.regex == '.*' or self.platform == 'windows':
                    print('\nWARNING: binary not found, skipping', binary)
        return sorted(out)

    def pre_build_steps(self):
        return []

    def build_steps(self):
        if self.platform == 'windows':
            return [[
                'tools\\run_tests\\helper_scripts\\build_cxx.bat',
                '-DgRPC_BUILD_MSVC_MP_COUNT=%d' % self.args.jobs
            ] + self._cmake_configure_extra_args]
        else:
            return [['tools/run_tests/helper_scripts/build_cxx.sh'] +
                    self._cmake_configure_extra_args]

    def build_steps_environ(self):
        """Extra environment variables set for pre_build_steps and build_steps jobs."""
        environ = {'GRPC_RUN_TESTS_CXX_LANGUAGE_SUFFIX': self.lang_suffix}
        if self.platform == 'windows':
            environ['GRPC_CMAKE_GENERATOR'] = self._cmake_generator_windows
            environ[
                'GRPC_CMAKE_ARCHITECTURE'] = self._cmake_architecture_windows
            environ[
                'GRPC_BUILD_ACTIVATE_VS_TOOLS'] = self._activate_vs_tools_windows
            environ[
                'GRPC_BUILD_VS_TOOLS_ARCHITECTURE'] = self._vs_tools_architecture_windows
        return environ

    def post_tests_steps(self):
        if self.platform == 'windows':
            return []
        else:
            return [['tools/run_tests/helper_scripts/post_tests_c.sh']]

    def _clang_cmake_configure_extra_args(self, version_suffix=''):
        return [
            '-DCMAKE_C_COMPILER=clang%s' % version_suffix,
            '-DCMAKE_CXX_COMPILER=clang++%s' % version_suffix,
        ]

    def _compiler_options(self, use_docker, compiler):
        """Returns docker distro and cmake configure args to use for given compiler."""
        if not use_docker and not _is_use_docker_child():
            # if not running under docker, we cannot ensure the right compiler version will be used,
            # so we only allow the non-specific choices.
            _check_compiler(compiler, ['default', 'cmake'])

        if compiler == 'default' or compiler == 'cmake':
            return ('debian11', [])
        elif compiler == 'gcc6':
            return ('gcc_6', [])
        elif compiler == 'gcc10.2':
            return ('debian11', [])
        elif compiler == 'gcc10.2_openssl102':
            return ('debian11_openssl102', [
                "-DgRPC_SSL_PROVIDER=package",
            ])
        elif compiler == 'gcc11':
            return ('gcc_11', [])
        elif compiler == 'gcc_musl':
            return ('alpine', [])
        elif compiler == 'clang6':
            return ('clang_6', self._clang_cmake_configure_extra_args())
        elif compiler == 'clang13':
            return ('clang_13', self._clang_cmake_configure_extra_args())
        else:
            raise Exception('Compiler %s not supported.' % compiler)

    def dockerfile_dir(self):
        return 'tools/dockerfile/test/cxx_%s_%s' % (
            self._docker_distro, _docker_arch_suffix(self.args.arch))

    def __str__(self):
        return self.lang_suffix


# This tests Node on grpc/grpc-node and will become the standard for Node testing
class RemoteNodeLanguage(object):

    def __init__(self):
        self.platform = platform_string()

    def configure(self, config, args):
        self.config = config
        self.args = args
        # Note: electron ABI only depends on major and minor version, so that's all
        # we should specify in the compiler argument
        _check_compiler(self.args.compiler, [
            'default', 'node0.12', 'node4', 'node5', 'node6', 'node7', 'node8',
            'electron1.3', 'electron1.6'
        ])
        if self.args.compiler == 'default':
            self.runtime = 'node'
            self.node_version = '8'
        else:
            if self.args.compiler.startswith('electron'):
                self.runtime = 'electron'
                self.node_version = self.args.compiler[8:]
            else:
                self.runtime = 'node'
                # Take off the word "node"
                self.node_version = self.args.compiler[4:]

    # TODO: update with Windows/electron scripts when available for grpc/grpc-node
    def test_specs(self):
        if self.platform == 'windows':
            return [
                self.config.job_spec(
                    ['tools\\run_tests\\helper_scripts\\run_node.bat'])
            ]
        else:
            return [
                self.config.job_spec(
                    ['tools/run_tests/helper_scripts/run_grpc-node.sh'],
                    None,
                    environ=_FORCE_ENVIRON_FOR_WRAPPERS)
            ]

    def pre_build_steps(self):
        return []

    def build_steps(self):
        return []

    def build_steps_environ(self):
        """Extra environment variables set for pre_build_steps and build_steps jobs."""
        return {}

    def post_tests_steps(self):
        return []

    def dockerfile_dir(self):
        return 'tools/dockerfile/test/node_jessie_%s' % _docker_arch_suffix(
            self.args.arch)

    def __str__(self):
        return 'grpc-node'


class Php7Language(object):

    def configure(self, config, args):
        self.config = config
        self.args = args
        _check_compiler(self.args.compiler, ['default'])

    def test_specs(self):
        return [
            self.config.job_spec(['src/php/bin/run_tests.sh'],
                                 environ=_FORCE_ENVIRON_FOR_WRAPPERS)
        ]

    def pre_build_steps(self):
        return []

    def build_steps(self):
        return [['tools/run_tests/helper_scripts/build_php.sh']]

    def build_steps_environ(self):
        """Extra environment variables set for pre_build_steps and build_steps jobs."""
        return {}

    def post_tests_steps(self):
        return [['tools/run_tests/helper_scripts/post_tests_php.sh']]

    def dockerfile_dir(self):
        return 'tools/dockerfile/test/php7_debian11_%s' % _docker_arch_suffix(
            self.args.arch)

    def __str__(self):
        return 'php7'


class PythonConfig(
        collections.namedtuple('PythonConfig', ['name', 'build', 'run'])):
    """Tuple of commands (named s.t. 'what it says on the tin' applies)"""


class PythonLanguage(object):

    _TEST_SPECS_FILE = {
        'native': ['src/python/grpcio_tests/tests/tests.json'],
        'gevent': [
            'src/python/grpcio_tests/tests/tests.json',
            'src/python/grpcio_tests/tests_gevent/tests.json',
        ],
        'asyncio': ['src/python/grpcio_tests/tests_aio/tests.json'],
    }

    _TEST_COMMAND = {
        'native': 'test_lite',
        'gevent': 'test_gevent',
        'asyncio': 'test_aio',
    }

    def configure(self, config, args):
        self.config = config
        self.args = args
        self.pythons = self._get_pythons(self.args)

    def test_specs(self):
        # load list of known test suites
        jobs = []
        for io_platform in self._TEST_SPECS_FILE:
            test_cases = []
            for tests_json_file_name in self._TEST_SPECS_FILE[io_platform]:
                with open(tests_json_file_name) as tests_json_file:
                    test_cases.extend(json.load(tests_json_file))

            environment = dict(_FORCE_ENVIRON_FOR_WRAPPERS)
            # TODO(https://github.com/grpc/grpc/issues/21401) Fork handlers is not
            # designed for non-native IO manager. It has a side-effect that
            # overrides threading settings in C-Core.
            if io_platform != 'native':
                environment['GRPC_ENABLE_FORK_SUPPORT'] = '0'
            for python_config in self.pythons:
                # TODO(https://github.com/grpc/grpc/issues/23784) allow gevent
                # to run on later version once issue solved.
                if io_platform == 'gevent' and python_config.name != 'py36':
                    continue
                jobs.extend([
                    self.config.job_spec(
                        python_config.run + [self._TEST_COMMAND[io_platform]],
                        timeout_seconds=8 * 60,
                        environ=dict(
                            GRPC_PYTHON_TESTRUNNER_FILTER=str(test_case),
                            **environment),
                        shortname='%s.%s.%s' %
                        (python_config.name, io_platform, test_case),
                    ) for test_case in test_cases
                ])
        return jobs

    def pre_build_steps(self):
        return []

    def build_steps(self):
        return [config.build for config in self.pythons]

    def build_steps_environ(self):
        """Extra environment variables set for pre_build_steps and build_steps jobs."""
        return {}

    def post_tests_steps(self):
        if self.config.build_config != 'gcov':
            return []
        else:
            return [['tools/run_tests/helper_scripts/post_tests_python.sh']]

    def dockerfile_dir(self):
        return 'tools/dockerfile/test/python_%s_%s' % (
            self._python_docker_distro_name(),
            _docker_arch_suffix(self.args.arch))

    def _python_docker_distro_name(self):
        """Choose the docker image to use based on python version."""
        if self.args.compiler == 'python_alpine':
            return 'alpine'
        else:
            return 'debian11_default'

    def _get_pythons(self, args):
        """Get python runtimes to test with, based on current platform, architecture, compiler etc."""
        if args.iomgr_platform != 'native':
            raise ValueError(
                'Python builds no longer differentiate IO Manager platforms, please use "native"'
            )

        if args.arch == 'x86':
            bits = '32'
        else:
            bits = '64'

        if os.name == 'nt':
            shell = ['bash']
            builder = [
                os.path.abspath(
                    'tools/run_tests/helper_scripts/build_python_msys2.sh')
            ]
            builder_prefix_arguments = ['MINGW{}'.format(bits)]
            venv_relative_python = ['Scripts/python.exe']
            toolchain = ['mingw32']
        else:
            shell = []
            builder = [
                os.path.abspath(
                    'tools/run_tests/helper_scripts/build_python.sh')
            ]
            builder_prefix_arguments = []
            venv_relative_python = ['bin/python']
            toolchain = ['unix']

        runner = [
            os.path.abspath('tools/run_tests/helper_scripts/run_python.sh')
        ]

        config_vars = _PythonConfigVars(shell, builder,
                                        builder_prefix_arguments,
                                        venv_relative_python, toolchain, runner)
        python36_config = _python_config_generator(name='py36',
                                                   major='3',
                                                   minor='6',
                                                   bits=bits,
                                                   config_vars=config_vars)
        python37_config = _python_config_generator(name='py37',
                                                   major='3',
                                                   minor='7',
                                                   bits=bits,
                                                   config_vars=config_vars)
        python38_config = _python_config_generator(name='py38',
                                                   major='3',
                                                   minor='8',
                                                   bits=bits,
                                                   config_vars=config_vars)
        python39_config = _python_config_generator(name='py39',
                                                   major='3',
                                                   minor='9',
                                                   bits=bits,
                                                   config_vars=config_vars)
        python310_config = _python_config_generator(name='py310',
                                                    major='3',
                                                    minor='10',
                                                    bits=bits,
                                                    config_vars=config_vars)
        pypy27_config = _pypy_config_generator(name='pypy',
                                               major='2',
                                               config_vars=config_vars)
        pypy32_config = _pypy_config_generator(name='pypy3',
                                               major='3',
                                               config_vars=config_vars)

        if args.compiler == 'default':
            if os.name == 'nt':
                return (python38_config,)
            elif os.uname()[0] == 'Darwin':
                # NOTE(rbellevi): Testing takes significantly longer on
                # MacOS, so we restrict the number of interpreter versions
                # tested.
                return (python38_config,)
            elif platform.machine() == 'aarch64':
                # Currently the python_debian11_default_arm64 docker image
                # only has python3.9 installed (and that seems sufficient
                # for arm64 testing)
                return (python39_config,)
            else:
                return (
                    python36_config,
                    python38_config,
                )
        elif args.compiler == 'python3.6':
            return (python36_config,)
        elif args.compiler == 'python3.7':
            return (python37_config,)
        elif args.compiler == 'python3.8':
            return (python38_config,)
        elif args.compiler == 'python3.9':
            return (python39_config,)
        elif args.compiler == 'python3.10':
            return (python310_config,)
        elif args.compiler == 'pypy':
            return (pypy27_config,)
        elif args.compiler == 'pypy3':
            return (pypy32_config,)
        elif args.compiler == 'python_alpine':
            return (python38_config,)
        elif args.compiler == 'all_the_cpythons':
            return (
                python36_config,
                python37_config,
                python38_config,
                python39_config,
                python310_config,
            )
        else:
            raise Exception('Compiler %s not supported.' % args.compiler)

    def __str__(self):
        return 'python'


class RubyLanguage(object):

    def configure(self, config, args):
        self.config = config
        self.args = args
        _check_compiler(self.args.compiler, ['default'])

    def test_specs(self):
        tests = [
            self.config.job_spec(['tools/run_tests/helper_scripts/run_ruby.sh'],
                                 timeout_seconds=10 * 60,
                                 environ=_FORCE_ENVIRON_FOR_WRAPPERS)
        ]
        for test in [
                'src/ruby/end2end/sig_handling_test.rb',
                'src/ruby/end2end/channel_state_test.rb',
                'src/ruby/end2end/channel_closing_test.rb',
                'src/ruby/end2end/sig_int_during_channel_watch_test.rb',
                'src/ruby/end2end/killed_client_thread_test.rb',
                'src/ruby/end2end/forking_client_test.rb',
                'src/ruby/end2end/grpc_class_init_test.rb',
                'src/ruby/end2end/multiple_killed_watching_threads_test.rb',
                'src/ruby/end2end/load_grpc_with_gc_stress_test.rb',
                'src/ruby/end2end/client_memory_usage_test.rb',
                'src/ruby/end2end/package_with_underscore_test.rb',
                'src/ruby/end2end/graceful_sig_handling_test.rb',
                'src/ruby/end2end/graceful_sig_stop_test.rb',
                'src/ruby/end2end/errors_load_before_grpc_lib_test.rb',
                'src/ruby/end2end/logger_load_before_grpc_lib_test.rb',
                'src/ruby/end2end/status_codes_load_before_grpc_lib_test.rb',
                'src/ruby/end2end/call_credentials_timeout_test.rb',
                'src/ruby/end2end/call_credentials_returning_bad_metadata_doesnt_kill_background_thread_test.rb'
        ]:
            tests.append(
                self.config.job_spec(['ruby', test],
                                     shortname=test,
                                     timeout_seconds=20 * 60,
                                     environ=_FORCE_ENVIRON_FOR_WRAPPERS))
        return tests

    def pre_build_steps(self):
        return [['tools/run_tests/helper_scripts/pre_build_ruby.sh']]

    def build_steps(self):
        return [['tools/run_tests/helper_scripts/build_ruby.sh']]

    def build_steps_environ(self):
        """Extra environment variables set for pre_build_steps and build_steps jobs."""
        return {}

    def post_tests_steps(self):
        return [['tools/run_tests/helper_scripts/post_tests_ruby.sh']]

    def dockerfile_dir(self):
        return 'tools/dockerfile/test/ruby_debian11_%s' % _docker_arch_suffix(
            self.args.arch)

    def __str__(self):
        return 'ruby'


class CSharpLanguage(object):

    def __init__(self):
        self.platform = platform_string()

    def configure(self, config, args):
        self.config = config
        self.args = args
        _check_compiler(self.args.compiler, ['default', 'coreclr', 'mono'])
        if self.args.compiler == 'default':
            # test both runtimes by default
            self.test_runtimes = ['coreclr', 'mono']
        else:
            # only test the specified runtime
            self.test_runtimes = [self.args.compiler]

        if self.platform == 'windows':
            _check_arch(self.args.arch, ['default'])
            self._cmake_arch_option = 'x64'
        else:
            self._docker_distro = 'debian11'

    def test_specs(self):
        with open('src/csharp/tests.json') as f:
            tests_by_assembly = json.load(f)

        msbuild_config = _MSBUILD_CONFIG[self.config.build_config]
        nunit_args = ['--labels=All', '--noresult', '--workers=1']

        specs = []
        for test_runtime in self.test_runtimes:
            if test_runtime == 'coreclr':
                assembly_extension = '.dll'
                assembly_subdir = 'bin/%s/netcoreapp3.1' % msbuild_config
                runtime_cmd = ['dotnet', 'exec']
            elif test_runtime == 'mono':
                assembly_extension = '.exe'
                assembly_subdir = 'bin/%s/net45' % msbuild_config
                if self.platform == 'windows':
                    runtime_cmd = []
                elif self.platform == 'mac':
                    # mono before version 5.2 on MacOS defaults to 32bit runtime
                    runtime_cmd = ['mono', '--arch=64']
                else:
                    runtime_cmd = ['mono']
            else:
                raise Exception('Illegal runtime "%s" was specified.')

            for assembly in six.iterkeys(tests_by_assembly):
                assembly_file = 'src/csharp/%s/%s/%s%s' % (
                    assembly, assembly_subdir, assembly, assembly_extension)

                # normally, run each test as a separate process
                for test in tests_by_assembly[assembly]:
                    cmdline = runtime_cmd + [assembly_file,
                                             '--test=%s' % test] + nunit_args
                    specs.append(
                        self.config.job_spec(
                            cmdline,
                            shortname='csharp.%s.%s' % (test_runtime, test),
                            environ=_FORCE_ENVIRON_FOR_WRAPPERS))
        return specs

    def pre_build_steps(self):
        if self.platform == 'windows':
            return [['tools\\run_tests\\helper_scripts\\pre_build_csharp.bat']]
        else:
            return [['tools/run_tests/helper_scripts/pre_build_csharp.sh']]

    def build_steps(self):
        if self.platform == 'windows':
            return [['tools\\run_tests\\helper_scripts\\build_csharp.bat']]
        else:
            return [['tools/run_tests/helper_scripts/build_csharp.sh']]

    def build_steps_environ(self):
        """Extra environment variables set for pre_build_steps and build_steps jobs."""
        if self.platform == 'windows':
            return {'ARCHITECTURE': self._cmake_arch_option}
        else:
            return {}

    def post_tests_steps(self):
        if self.platform == 'windows':
            return [['tools\\run_tests\\helper_scripts\\post_tests_csharp.bat']]
        else:
            return [['tools/run_tests/helper_scripts/post_tests_csharp.sh']]

    def dockerfile_dir(self):
        return 'tools/dockerfile/test/csharp_%s_%s' % (
            self._docker_distro, _docker_arch_suffix(self.args.arch))

    def __str__(self):
        return 'csharp'


class ObjCLanguage(object):

    def configure(self, config, args):
        self.config = config
        self.args = args
        _check_compiler(self.args.compiler, ['default'])

    def test_specs(self):
        out = []
        # Currently not supporting compiling as frameworks in Bazel
        # TODO(jtattermusch): verify the above claim is still accurate.
        out.append(
            self.config.job_spec(
                ['src/objective-c/tests/build_one_example.sh'],
                timeout_seconds=20 * 60,
                shortname='ios-buildtest-example-sample-frameworks',
                cpu_cost=1e6,
                environ={
                    'SCHEME': 'Sample',
                    'EXAMPLE_PATH': 'src/objective-c/examples/Sample',
                    'FRAMEWORKS': 'YES'
                }))
        # TODO(jtattermusch): Create bazel target for the sample and remove the test task from here.
        out.append(
            self.config.job_spec(
                ['src/objective-c/tests/build_one_example.sh'],
                timeout_seconds=20 * 60,
                shortname='ios-buildtest-example-switftsample',
                cpu_cost=1e6,
                environ={
                    'SCHEME': 'SwiftSample',
                    'EXAMPLE_PATH': 'src/objective-c/examples/SwiftSample'
                }))
        # Disabled due to #20258
        # TODO (mxyan): Reenable this test when #20258 is resolved.
        # out.append(
        #     self.config.job_spec(
        #         ['src/objective-c/tests/build_one_example_bazel.sh'],
        #         timeout_seconds=20 * 60,
        #         shortname='ios-buildtest-example-watchOS-sample',
        #         cpu_cost=1e6,
        #         environ={
        #             'SCHEME': 'watchOS-sample-WatchKit-App',
        #             'EXAMPLE_PATH': 'src/objective-c/examples/watchOS-sample',
        #             'FRAMEWORKS': 'NO'
        #         }))

        # TODO(jtattermusch): move the test out of the test/core/iomgr/CFStreamTests directory?
        # How does one add the cfstream dependency in bazel?
        out.append(
            self.config.job_spec(
                ['test/core/iomgr/ios/CFStreamTests/build_and_run_tests.sh'],
                timeout_seconds=60 * 60,
                shortname='ios-test-cfstream-tests',
                cpu_cost=1e6,
                environ=_FORCE_ENVIRON_FOR_WRAPPERS))
        # TODO(jtattermusch): Create bazel target for the test and remove the test from here
        # (how does one add the cronet dependency in bazel?)
        out.append(
            self.config.job_spec(['src/objective-c/tests/run_one_test.sh'],
                                 timeout_seconds=60 * 60,
                                 shortname='ios-test-cronettests',
                                 cpu_cost=1e6,
                                 environ={'SCHEME': 'CronetTests'}))
        # TODO(jtattermusch): Create bazel target for the test and remove the test from here.
        out.append(
            self.config.job_spec(['src/objective-c/tests/run_one_test.sh'],
                                 timeout_seconds=30 * 60,
                                 shortname='ios-perf-test',
                                 cpu_cost=1e6,
                                 environ={'SCHEME': 'PerfTests'}))
        # TODO(jtattermusch): Clarify what's the difference between PerfTests and PerfTestsPosix
        # TODO(jtattermusch): Create bazel target for the test and remove the test from here.
        out.append(
            self.config.job_spec(['src/objective-c/tests/run_one_test.sh'],
                                 timeout_seconds=30 * 60,
                                 shortname='ios-perf-test-posix',
                                 cpu_cost=1e6,
                                 environ={'SCHEME': 'PerfTestsPosix'}))
        # TODO(jtattermusch): Create bazel target for the test (how does one add the cronet dependency in bazel?)
        # TODO(jtattermusch): move the test out of the test/cpp/ios directory?
        out.append(
            self.config.job_spec(['test/cpp/ios/build_and_run_tests.sh'],
                                 timeout_seconds=60 * 60,
                                 shortname='ios-cpp-test-cronet',
                                 cpu_cost=1e6,
                                 environ=_FORCE_ENVIRON_FOR_WRAPPERS))
        # TODO(jtattermusch): Make sure the //src/objective-c/tests:TvTests bazel test passes and remove the test from here.
        out.append(
            self.config.job_spec(['src/objective-c/tests/run_one_test.sh'],
                                 timeout_seconds=30 * 60,
                                 shortname='tvos-test-basictests',
                                 cpu_cost=1e6,
                                 environ={
                                     'SCHEME': 'TvTests',
                                     'PLATFORM': 'tvos'
                                 }))

        return sorted(out)

    def pre_build_steps(self):
        return []

    def build_steps(self):
        return []

    def build_steps_environ(self):
        """Extra environment variables set for pre_build_steps and build_steps jobs."""
        return {}

    def post_tests_steps(self):
        return []

    def dockerfile_dir(self):
        return None

    def __str__(self):
        return 'objc'


class Sanity(object):

    def configure(self, config, args):
        self.config = config
        self.args = args
        _check_compiler(self.args.compiler, ['default'])

    def test_specs(self):
        import yaml
        with open('tools/run_tests/sanity/sanity_tests.yaml', 'r') as f:
            environ = {'TEST': 'true'}
            if _is_use_docker_child():
                environ['CLANG_FORMAT_SKIP_DOCKER'] = 'true'
                environ['CLANG_TIDY_SKIP_DOCKER'] = 'true'
                environ['IWYU_SKIP_DOCKER'] = 'true'
                # sanity tests run tools/bazel wrapper concurrently
                # and that can result in a download/run race in the wrapper.
                # under docker we already have the right version of bazel
                # so we can just disable the wrapper.
                environ['DISABLE_BAZEL_WRAPPER'] = 'true'
            return [
                self.config.job_spec(cmd['script'].split(),
                                     timeout_seconds=30 * 60,
                                     environ=environ,
                                     cpu_cost=cmd.get('cpu_cost', 1))
                for cmd in yaml.load(f)
            ]

    def pre_build_steps(self):
        return []

    def build_steps(self):
        return []

    def build_steps_environ(self):
        """Extra environment variables set for pre_build_steps and build_steps jobs."""
        return {}

    def post_tests_steps(self):
        return []

    def dockerfile_dir(self):
        return 'tools/dockerfile/test/sanity'

    def __str__(self):
        return 'sanity'


# different configurations we can run under
with open('tools/run_tests/generated/configs.json') as f:
    _CONFIGS = dict(
        (cfg['config'], Config(**cfg)) for cfg in ast.literal_eval(f.read()))

_LANGUAGES = {
    'c++': CLanguage('cxx', 'c++'),
    'c': CLanguage('c', 'c'),
    'grpc-node': RemoteNodeLanguage(),
    'php7': Php7Language(),
    'python': PythonLanguage(),
    'ruby': RubyLanguage(),
    'csharp': CSharpLanguage(),
    'objc': ObjCLanguage(),
    'sanity': Sanity()
}

_MSBUILD_CONFIG = {
    'dbg': 'Debug',
    'opt': 'Release',
    'gcov': 'Debug',
}


def _build_step_environ(cfg, extra_env={}):
    """Environment variables set for each build step."""
    environ = {'CONFIG': cfg, 'GRPC_RUN_TESTS_JOBS': str(args.jobs)}
    msbuild_cfg = _MSBUILD_CONFIG.get(cfg)
    if msbuild_cfg:
        environ['MSBUILD_CONFIG'] = msbuild_cfg
    environ.update(extra_env)
    return environ


def _windows_arch_option(arch):
    """Returns msbuild cmdline option for selected architecture."""
    if arch == 'default' or arch == 'x86':
        return '/p:Platform=Win32'
    elif arch == 'x64':
        return '/p:Platform=x64'
    else:
        print('Architecture %s not supported.' % arch)
        sys.exit(1)


def _check_arch_option(arch):
    """Checks that architecture option is valid."""
    if platform_string() == 'windows':
        _windows_arch_option(arch)
    elif platform_string() == 'linux':
        # On linux, we need to be running under docker with the right architecture.
        runtime_machine = platform.machine()
        runtime_arch = platform.architecture()[0]
        if arch == 'default':
            return
        elif runtime_machine == 'x86_64' and runtime_arch == '64bit' and arch == 'x64':
            return
        elif runtime_machine == 'x86_64' and runtime_arch == '32bit' and arch == 'x86':
            return
        elif runtime_machine == 'aarch64' and runtime_arch == '64bit' and arch == 'arm64':
            return
        else:
            print(
                'Architecture %s does not match current runtime architecture.' %
                arch)
            sys.exit(1)
    else:
        if args.arch != 'default':
            print('Architecture %s not supported on current platform.' %
                  args.arch)
            sys.exit(1)


def _docker_arch_suffix(arch):
    """Returns suffix to dockerfile dir to use."""
    if arch == 'default' or arch == 'x64':
        return 'x64'
    elif arch == 'x86':
        return 'x86'
    elif arch == 'arm64':
        return 'arm64'
    else:
        print('Architecture %s not supported with current settings.' % arch)
        sys.exit(1)


def runs_per_test_type(arg_str):
    """Auxiliary function to parse the "runs_per_test" flag.

       Returns:
           A positive integer or 0, the latter indicating an infinite number of
           runs.

       Raises:
           argparse.ArgumentTypeError: Upon invalid input.
    """
    if arg_str == 'inf':
        return 0
    try:
        n = int(arg_str)
        if n <= 0:
            raise ValueError
        return n
    except:
        msg = '\'{}\' is not a positive integer or \'inf\''.format(arg_str)
        raise argparse.ArgumentTypeError(msg)


def percent_type(arg_str):
    pct = float(arg_str)
    if pct > 100 or pct < 0:
        raise argparse.ArgumentTypeError(
            "'%f' is not a valid percentage in the [0, 100] range" % pct)
    return pct


# This is math.isclose in python >= 3.5
def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def _shut_down_legacy_server(legacy_server_port):
    """Shut down legacy version of port server."""
    try:
        version = int(
            urllib.request.urlopen('http://localhost:%d/version_number' %
                                   legacy_server_port,
                                   timeout=10).read())
    except:
        pass
    else:
        urllib.request.urlopen('http://localhost:%d/quitquitquit' %
                               legacy_server_port).read()


def _calculate_num_runs_failures(list_of_results):
    """Calculate number of runs and failures for a particular test.

  Args:
    list_of_results: (List) of JobResult object.
  Returns:
    A tuple of total number of runs and failures.
  """
    num_runs = len(list_of_results)  # By default, there is 1 run per JobResult.
    num_failures = 0
    for jobresult in list_of_results:
        if jobresult.retries > 0:
            num_runs += jobresult.retries
        if jobresult.num_failures > 0:
            num_failures += jobresult.num_failures
    return num_runs, num_failures


class BuildAndRunError(object):
    """Represents error type in _build_and_run."""

    BUILD = object()
    TEST = object()
    POST_TEST = object()


# returns a list of things that failed (or an empty list on success)
def _build_and_run(check_cancelled,
                   newline_on_success,
                   xml_report=None,
                   build_only=False):
    """Do one pass of building & running tests."""
    # build latest sequentially
    num_failures, resultset = jobset.run(build_steps,
                                         maxjobs=1,
                                         stop_on_failure=True,
                                         newline_on_success=newline_on_success,
                                         travis=args.travis)
    if num_failures:
        return [BuildAndRunError.BUILD]

    if build_only:
        if xml_report:
            report_utils.render_junit_xml_report(
                resultset, xml_report, suite_name=args.report_suite_name)
        return []

    # start antagonists
    antagonists = [
        subprocess.Popen(['tools/run_tests/python_utils/antagonist.py'])
        for _ in range(0, args.antagonists)
    ]
    start_port_server.start_port_server()
    resultset = None
    num_test_failures = 0
    try:
        infinite_runs = runs_per_test == 0
        one_run = set(spec for language in languages
                      for spec in language.test_specs()
                      if (re.search(args.regex, spec.shortname) and
                          (args.regex_exclude == '' or
                           not re.search(args.regex_exclude, spec.shortname))))
        # When running on travis, we want out test runs to be as similar as possible
        # for reproducibility purposes.
        if args.travis and args.max_time <= 0:
            massaged_one_run = sorted(one_run, key=lambda x: x.cpu_cost)
        else:
            # whereas otherwise, we want to shuffle things up to give all tests a
            # chance to run.
            massaged_one_run = list(
                one_run)  # random.sample needs an indexable seq.
            num_jobs = len(massaged_one_run)
            # for a random sample, get as many as indicated by the 'sample_percent'
            # argument. By default this arg is 100, resulting in a shuffle of all
            # jobs.
            sample_size = int(num_jobs * args.sample_percent / 100.0)
            massaged_one_run = random.sample(massaged_one_run, sample_size)
            if not isclose(args.sample_percent, 100.0):
                assert args.runs_per_test == 1, "Can't do sampling (-p) over multiple runs (-n)."
                print("Running %d tests out of %d (~%d%%)" %
                      (sample_size, num_jobs, args.sample_percent))
        if infinite_runs:
            assert len(massaged_one_run
                      ) > 0, 'Must have at least one test for a -n inf run'
        runs_sequence = (itertools.repeat(massaged_one_run) if infinite_runs
                         else itertools.repeat(massaged_one_run, runs_per_test))
        all_runs = itertools.chain.from_iterable(runs_sequence)

        if args.quiet_success:
            jobset.message(
                'START',
                'Running tests quietly, only failing tests will be reported',
                do_newline=True)
        num_test_failures, resultset = jobset.run(
            all_runs,
            check_cancelled,
            newline_on_success=newline_on_success,
            travis=args.travis,
            maxjobs=args.jobs,
            maxjobs_cpu_agnostic=max_parallel_tests_for_current_platform(),
            stop_on_failure=args.stop_on_failure,
            quiet_success=args.quiet_success,
            max_time=args.max_time)
        if resultset:
            for k, v in sorted(resultset.items()):
                num_runs, num_failures = _calculate_num_runs_failures(v)
                if num_failures > 0:
                    if num_failures == num_runs:  # what about infinite_runs???
                        jobset.message('FAILED', k, do_newline=True)
                    else:
                        jobset.message('FLAKE',
                                       '%s [%d/%d runs flaked]' %
                                       (k, num_failures, num_runs),
                                       do_newline=True)
    finally:
        for antagonist in antagonists:
            antagonist.kill()
        if args.bq_result_table and resultset:
            upload_extra_fields = {
                'compiler': args.compiler,
                'config': args.config,
                'iomgr_platform': args.iomgr_platform,
                'language': args.language[
                    0
                ],  # args.language is a list but will always have one element when uploading to BQ is enabled.
                'platform': platform_string()
            }
            try:
                upload_results_to_bq(resultset, args.bq_result_table,
                                     upload_extra_fields)
            except NameError as e:
                logging.warning(
                    e)  # It's fine to ignore since this is not critical
        if xml_report and resultset:
            report_utils.render_junit_xml_report(
                resultset,
                xml_report,
                suite_name=args.report_suite_name,
                multi_target=args.report_multi_target)

    number_failures, _ = jobset.run(post_tests_steps,
                                    maxjobs=1,
                                    stop_on_failure=False,
                                    newline_on_success=newline_on_success,
                                    travis=args.travis)

    out = []
    if number_failures:
        out.append(BuildAndRunError.POST_TEST)
    if num_test_failures:
        out.append(BuildAndRunError.TEST)

    return out


# parse command line
argp = argparse.ArgumentParser(description='Run grpc tests.')
argp.add_argument('-c',
                  '--config',
                  choices=sorted(_CONFIGS.keys()),
                  default='opt')
argp.add_argument(
    '-n',
    '--runs_per_test',
    default=1,
    type=runs_per_test_type,
    help='A positive integer or "inf". If "inf", all tests will run in an '
    'infinite loop. Especially useful in combination with "-f"')
argp.add_argument('-r', '--regex', default='.*', type=str)
argp.add_argument('--regex_exclude', default='', type=str)
argp.add_argument('-j', '--jobs', default=multiprocessing.cpu_count(), type=int)
argp.add_argument('-s', '--slowdown', default=1.0, type=float)
argp.add_argument('-p',
                  '--sample_percent',
                  default=100.0,
                  type=percent_type,
                  help='Run a random sample with that percentage of tests')
argp.add_argument(
    '-t',
    '--travis',
    default=False,
    action='store_const',
    const=True,
    help='When set, indicates that the script is running on CI (= not locally).'
)
argp.add_argument('--newline_on_success',
                  default=False,
                  action='store_const',
                  const=True)
argp.add_argument('-l',
                  '--language',
                  choices=sorted(_LANGUAGES.keys()),
                  nargs='+',
                  required=True)
argp.add_argument('-S',
                  '--stop_on_failure',
                  default=False,
                  action='store_const',
                  const=True)
argp.add_argument('--use_docker',
                  default=False,
                  action='store_const',
                  const=True,
                  help='Run all the tests under docker. That provides ' +
                  'additional isolation and prevents the need to install ' +
                  'language specific prerequisites. Only available on Linux.')
argp.add_argument(
    '--allow_flakes',
    default=False,
    action='store_const',
    const=True,
    help=
    'Allow flaky tests to show as passing (re-runs failed tests up to five times)'
)
argp.add_argument(
    '--arch',
    choices=['default', 'x86', 'x64', 'arm64'],
    default='default',
    help=
    'Selects architecture to target. For some platforms "default" is the only supported choice.'
)
argp.add_argument(
    '--compiler',
    choices=[
        'default',
        'gcc6',
        'gcc10.2',
        'gcc10.2_openssl102',
        'gcc11',
        'gcc_musl',
        'clang6',
        'clang13',
        'python2.7',
        'python3.5',
        'python3.6',
        'python3.7',
        'python3.8',
        'python3.9',
        'pypy',
        'pypy3',
        'python_alpine',
        'all_the_cpythons',
        'electron1.3',
        'electron1.6',
        'coreclr',
        'cmake',
        'cmake_ninja_vs2017',
        'cmake_vs2017',
        'cmake_vs2019',
        'mono',
    ],
    default='default',
    help=
    'Selects compiler to use. Allowed values depend on the platform and language.'
)
argp.add_argument('--iomgr_platform',
                  choices=['native', 'gevent', 'asyncio'],
                  default='native',
                  help='Selects iomgr platform to build on')
argp.add_argument('--build_only',
                  default=False,
                  action='store_const',
                  const=True,
                  help='Perform all the build steps but don\'t run any tests.')
argp.add_argument('--measure_cpu_costs',
                  default=False,
                  action='store_const',
                  const=True,
                  help='Measure the cpu costs of tests')
argp.add_argument('-a', '--antagonists', default=0, type=int)
argp.add_argument('-x',
                  '--xml_report',
                  default=None,
                  type=str,
                  help='Generates a JUnit-compatible XML report')
argp.add_argument('--report_suite_name',
                  default='tests',
                  type=str,
                  help='Test suite name to use in generated JUnit XML report')
argp.add_argument(
    '--report_multi_target',
    default=False,
    const=True,
    action='store_const',
    help='Generate separate XML report for each test job (Looks better in UIs).'
)
argp.add_argument(
    '--quiet_success',
    default=False,
    action='store_const',
    const=True,
    help=
    'Don\'t print anything when a test passes. Passing tests also will not be reported in XML report. '
    + 'Useful when running many iterations of each test (argument -n).')
argp.add_argument(
    '--force_default_poller',
    default=False,
    action='store_const',
    const=True,
    help='Don\'t try to iterate over many polling strategies when they exist')
argp.add_argument(
    '--force_use_pollers',
    default=None,
    type=str,
    help='Only use the specified comma-delimited list of polling engines. '
    'Example: --force_use_pollers epoll1,poll '
    ' (This flag has no effect if --force_default_poller flag is also used)')
argp.add_argument('--max_time',
                  default=-1,
                  type=int,
                  help='Maximum test runtime in seconds')
argp.add_argument('--bq_result_table',
                  default='',
                  type=str,
                  nargs='?',
                  help='Upload test results to a specified BQ table.')
args = argp.parse_args()

flaky_tests = set()
shortname_to_cpu = {}

if args.force_default_poller:
    _POLLING_STRATEGIES = {}
elif args.force_use_pollers:
    _POLLING_STRATEGIES[platform_string()] = args.force_use_pollers.split(',')

jobset.measure_cpu_costs = args.measure_cpu_costs

# grab config
run_config = _CONFIGS[args.config]
build_config = run_config.build_config

# TODO(jtattermusch): is this setting applied/being used?
if args.travis:
    _FORCE_ENVIRON_FOR_WRAPPERS = {'GRPC_TRACE': 'api'}

languages = set(_LANGUAGES[l] for l in args.language)
for l in languages:
    l.configure(run_config, args)

if len(languages) != 1:
    print('Building multiple languages simultaneously is not supported!')
    sys.exit(1)

# If --use_docker was used, respawn the run_tests.py script under a docker container
# instead of continuing.
if args.use_docker:
    if not args.travis:
        print('Seen --use_docker flag, will run tests under docker.')
        print('')
        print(
            'IMPORTANT: The changes you are testing need to be locally committed'
        )
        print(
            'because only the committed changes in the current branch will be')
        print('copied to the docker environment.')
        time.sleep(5)

    dockerfile_dirs = set([l.dockerfile_dir() for l in languages])
    if len(dockerfile_dirs) > 1:
        print('Languages to be tested require running under different docker '
              'images.')
        sys.exit(1)
    else:
        dockerfile_dir = next(iter(dockerfile_dirs))

    child_argv = [arg for arg in sys.argv if not arg == '--use_docker']
    run_tests_cmd = 'python3 tools/run_tests/run_tests.py %s' % ' '.join(
        child_argv[1:])

    env = os.environ.copy()
    env['DOCKERFILE_DIR'] = dockerfile_dir
    env['DOCKER_RUN_SCRIPT'] = 'tools/run_tests/dockerize/docker_run.sh'
    env['DOCKER_RUN_SCRIPT_COMMAND'] = run_tests_cmd

    retcode = subprocess.call(
        'tools/run_tests/dockerize/build_and_run_docker.sh',
        shell=True,
        env=env)
    _print_debug_info_epilogue(dockerfile_dir=dockerfile_dir)
    sys.exit(retcode)

_check_arch_option(args.arch)

# collect pre-build steps (which get retried if they fail, e.g. to avoid
# flakes on downloading dependencies etc.)
build_steps = list(
    set(
        jobset.JobSpec(cmdline,
                       environ=_build_step_environ(
                           build_config, extra_env=l.build_steps_environ()),
                       timeout_seconds=_PRE_BUILD_STEP_TIMEOUT_SECONDS,
                       flake_retries=2)
        for l in languages
        for cmdline in l.pre_build_steps()))

# collect build steps
build_steps.extend(
    set(
        jobset.JobSpec(cmdline,
                       environ=_build_step_environ(
                           build_config, extra_env=l.build_steps_environ()),
                       timeout_seconds=None)
        for l in languages
        for cmdline in l.build_steps()))

# collect post test steps
post_tests_steps = list(
    set(
        jobset.JobSpec(cmdline,
                       environ=_build_step_environ(
                           build_config, extra_env=l.build_steps_environ()))
        for l in languages
        for cmdline in l.post_tests_steps()))
runs_per_test = args.runs_per_test

errors = _build_and_run(check_cancelled=lambda: False,
                        newline_on_success=args.newline_on_success,
                        xml_report=args.xml_report,
                        build_only=args.build_only)
if not errors:
    jobset.message('SUCCESS', 'All tests passed', do_newline=True)
else:
    jobset.message('FAILED', 'Some tests failed', do_newline=True)

if not _is_use_docker_child():
    # if --use_docker was used, the outer invocation of run_tests.py will
    # print the debug info instead.
    _print_debug_info_epilogue()

exit_code = 0
if BuildAndRunError.BUILD in errors:
    exit_code |= 1
if BuildAndRunError.TEST in errors:
    exit_code |= 2
if BuildAndRunError.POST_TEST in errors:
    exit_code |= 4
sys.exit(exit_code)



# Copyright 2021 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import List

from absl import flags
from absl.testing import absltest

from framework import xds_k8s_testcase
from framework.infrastructure import k8s
from framework.test_app import server_app

logger = logging.getLogger(__name__)
flags.adopt_module_key_flags(xds_k8s_testcase)

# Type aliases
_XdsTestServer = xds_k8s_testcase.XdsTestServer
_XdsTestClient = xds_k8s_testcase.XdsTestClient


class FailoverTest(xds_k8s_testcase.RegularXdsKubernetesTestCase):
    REPLICA_COUNT = 3
    MAX_RATE_PER_ENDPOINT = 100

    def setUp(self):
        super().setUp()
        self.secondary_server_runner = server_app.KubernetesServerRunner(
            k8s.KubernetesNamespace(self.secondary_k8s_api_manager,
                                    self.server_namespace),
            deployment_name=self.server_name + '-alt',
            image_name=self.server_image,
            gcp_service_account=self.gcp_service_account,
            td_bootstrap_image=self.td_bootstrap_image,
            gcp_project=self.project,
            gcp_api_manager=self.gcp_api_manager,
            xds_server_uri=self.xds_server_uri,
            network=self.network,
            debug_use_port_forwarding=self.debug_use_port_forwarding,
            reuse_namespace=True)

    def tearDown(self):
        if hasattr(self, 'secondary_server_runner'):
            self.secondary_server_runner.cleanup()
        super().tearDown()

    def test_failover(self) -> None:
        with self.subTest('00_create_health_check'):
            self.td.create_health_check()

        with self.subTest('01_create_backend_services'):
            self.td.create_backend_service()

        with self.subTest('02_create_url_map'):
            self.td.create_url_map(self.server_xds_host, self.server_xds_port)

        with self.subTest('03_create_target_proxy'):
            self.td.create_target_proxy()

        with self.subTest('04_create_forwarding_rule'):
            self.td.create_forwarding_rule(self.server_xds_port)

        default_test_servers: List[_XdsTestServer]
        alternate_test_servers: List[_XdsTestServer]
        with self.subTest('05_start_test_servers'):
            default_test_servers = self.startTestServers(
                replica_count=self.REPLICA_COUNT)

            alternate_test_servers = self.startTestServers(
                server_runner=self.secondary_server_runner)

        with self.subTest('06_add_server_backends_to_backend_services'):
            self.setupServerBackends(
                max_rate_per_endpoint=self.MAX_RATE_PER_ENDPOINT)
            self.setupServerBackends(
                server_runner=self.secondary_server_runner,
                max_rate_per_endpoint=self.MAX_RATE_PER_ENDPOINT)

        test_client: _XdsTestClient
        with self.subTest('07_start_test_client'):
            test_client = self.startTestClient(default_test_servers[0])

        with self.subTest('08_test_client_xds_config_exists'):
            self.assertXdsConfigExists(test_client)

        with self.subTest('09_primary_locality_receives_requests'):
            self.assertRpcsEventuallyGoToGivenServers(test_client,
                                                      default_test_servers)

        with self.subTest(
                '10_secondary_locality_receives_no_requests_on_partial_primary_failure'
        ):
            default_test_servers[0].set_not_serving()
            self.assertRpcsEventuallyGoToGivenServers(test_client,
                                                      default_test_servers[1:])

        with self.subTest('11_gentle_failover'):
            default_test_servers[1].set_not_serving()
            self.assertRpcsEventuallyGoToGivenServers(
                test_client, default_test_servers[2:] + alternate_test_servers)

        with self.subTest(
                '12_secondary_locality_receives_requests_on_primary_failure'):
            default_test_servers[2].set_not_serving()
            self.assertRpcsEventuallyGoToGivenServers(test_client,
                                                      alternate_test_servers)

        with self.subTest('13_traffic_resumes_to_healthy_backends'):
            for i in range(self.REPLICA_COUNT):
                default_test_servers[i].set_serving()
            self.assertRpcsEventuallyGoToGivenServers(test_client,
                                                      default_test_servers)


if __name__ == '__main__':
    absltest.main(failfast=True)



# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Governs coverage for tests of RPCs throughout RPC Framework."""

import abc

import six

# This code is designed for use with the unittest module.
# pylint: disable=invalid-name


class Coverage(six.with_metaclass(abc.ABCMeta)):
    """Specification of test coverage."""

    @abc.abstractmethod
    def testSuccessfulUnaryRequestUnaryResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testSuccessfulUnaryRequestStreamResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testSuccessfulStreamRequestUnaryResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testSuccessfulStreamRequestStreamResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testSequentialInvocations(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testParallelInvocations(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testWaitingForSomeButNotAllParallelInvocations(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testCancelledUnaryRequestUnaryResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testCancelledUnaryRequestStreamResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testCancelledStreamRequestUnaryResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testCancelledStreamRequestStreamResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testExpiredUnaryRequestUnaryResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testExpiredUnaryRequestStreamResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testExpiredStreamRequestUnaryResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testExpiredStreamRequestStreamResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testFailedUnaryRequestUnaryResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testFailedUnaryRequestStreamResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testFailedStreamRequestUnaryResponse(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def testFailedStreamRequestStreamResponse(self):
        raise NotImplementedError()



# Copyright 2019 The gRPC Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import time
import unittest

import grpc
from grpc.experimental import aio

from tests.unit import resources
from tests.unit.framework.common import test_constants
from tests_aio.unit._test_base import AioTestBase

_SIMPLE_UNARY_UNARY = '/test/SimpleUnaryUnary'
_BLOCK_FOREVER = '/test/BlockForever'
_BLOCK_BRIEFLY = '/test/BlockBriefly'
_UNARY_STREAM_ASYNC_GEN = '/test/UnaryStreamAsyncGen'
_UNARY_STREAM_READER_WRITER = '/test/UnaryStreamReaderWriter'
_UNARY_STREAM_EVILLY_MIXED = '/test/UnaryStreamEvillyMixed'
_STREAM_UNARY_ASYNC_GEN = '/test/StreamUnaryAsyncGen'
_STREAM_UNARY_READER_WRITER = '/test/StreamUnaryReaderWriter'
_STREAM_UNARY_EVILLY_MIXED = '/test/StreamUnaryEvillyMixed'
_STREAM_STREAM_ASYNC_GEN = '/test/StreamStreamAsyncGen'
_STREAM_STREAM_READER_WRITER = '/test/StreamStreamReaderWriter'
_STREAM_STREAM_EVILLY_MIXED = '/test/StreamStreamEvillyMixed'
_UNIMPLEMENTED_METHOD = '/test/UnimplementedMethod'
_ERROR_IN_STREAM_STREAM = '/test/ErrorInStreamStream'
_ERROR_IN_STREAM_UNARY = '/test/ErrorInStreamUnary'
_ERROR_WITHOUT_RAISE_IN_UNARY_UNARY = '/test/ErrorWithoutRaiseInUnaryUnary'
_ERROR_WITHOUT_RAISE_IN_STREAM_STREAM = '/test/ErrorWithoutRaiseInStreamStream'
_INVALID_TRAILING_METADATA = '/test/InvalidTrailingMetadata'

_REQUEST = b'\x00\x00\x00'
_RESPONSE = b'\x01\x01\x01'
_NUM_STREAM_REQUESTS = 3
_NUM_STREAM_RESPONSES = 5
_MAXIMUM_CONCURRENT_RPCS = 5


class _GenericHandler(grpc.GenericRpcHandler):

    def __init__(self):
        self._called = asyncio.get_event_loop().create_future()
        self._routing_table = {
            _SIMPLE_UNARY_UNARY:
                grpc.unary_unary_rpc_method_handler(self._unary_unary),
            _BLOCK_FOREVER:
                grpc.unary_unary_rpc_method_handler(self._block_forever),
            _BLOCK_BRIEFLY:
                grpc.unary_unary_rpc_method_handler(self._block_briefly),
            _UNARY_STREAM_ASYNC_GEN:
                grpc.unary_stream_rpc_method_handler(
                    self._unary_stream_async_gen),
            _UNARY_STREAM_READER_WRITER:
                grpc.unary_stream_rpc_method_handler(
                    self._unary_stream_reader_writer),
            _UNARY_STREAM_EVILLY_MIXED:
                grpc.unary_stream_rpc_method_handler(
                    self._unary_stream_evilly_mixed),
            _STREAM_UNARY_ASYNC_GEN:
                grpc.stream_unary_rpc_method_handler(
                    self._stream_unary_async_gen),
            _STREAM_UNARY_READER_WRITER:
                grpc.stream_unary_rpc_method_handler(
                    self._stream_unary_reader_writer),
            _STREAM_UNARY_EVILLY_MIXED:
                grpc.stream_unary_rpc_method_handler(
                    self._stream_unary_evilly_mixed),
            _STREAM_STREAM_ASYNC_GEN:
                grpc.stream_stream_rpc_method_handler(
                    self._stream_stream_async_gen),
            _STREAM_STREAM_READER_WRITER:
                grpc.stream_stream_rpc_method_handler(
                    self._stream_stream_reader_writer),
            _STREAM_STREAM_EVILLY_MIXED:
                grpc.stream_stream_rpc_method_handler(
                    self._stream_stream_evilly_mixed),
            _ERROR_IN_STREAM_STREAM:
                grpc.stream_stream_rpc_method_handler(
                    self._error_in_stream_stream),
            _ERROR_IN_STREAM_UNARY:
                grpc.stream_unary_rpc_method_handler(
                    self._value_error_in_stream_unary),
            _ERROR_WITHOUT_RAISE_IN_UNARY_UNARY:
                grpc.unary_unary_rpc_method_handler(
                    self._error_without_raise_in_unary_unary),
            _ERROR_WITHOUT_RAISE_IN_STREAM_STREAM:
                grpc.stream_stream_rpc_method_handler(
                    self._error_without_raise_in_stream_stream),
            _INVALID_TRAILING_METADATA:
                grpc.unary_unary_rpc_method_handler(
                    self._invalid_trailing_metadata),
        }

    @staticmethod
    async def _unary_unary(unused_request, unused_context):
        return _RESPONSE

    async def _block_forever(self, unused_request, unused_context):
        await asyncio.get_event_loop().create_future()

    async def _block_briefly(self, unused_request, unused_context):
        await asyncio.sleep(test_constants.SHORT_TIMEOUT / 2)
        return _RESPONSE

    async def _unary_stream_async_gen(self, unused_request, unused_context):
        for _ in range(_NUM_STREAM_RESPONSES):
            yield _RESPONSE

    async def _unary_stream_reader_writer(self, unused_request, context):
        for _ in range(_NUM_STREAM_RESPONSES):
            await context.write(_RESPONSE)

    async def _unary_stream_evilly_mixed(self, unused_request, context):
        yield _RESPONSE
        for _ in range(_NUM_STREAM_RESPONSES - 1):
            await context.write(_RESPONSE)

    async def _stream_unary_async_gen(self, request_iterator, unused_context):
        request_count = 0
        async for request in request_iterator:
            assert _REQUEST == request
            request_count += 1
        assert _NUM_STREAM_REQUESTS == request_count
        return _RESPONSE

    async def _stream_unary_reader_writer(self, unused_request, context):
        for _ in range(_NUM_STREAM_REQUESTS):
            assert _REQUEST == await context.read()
        return _RESPONSE

    async def _stream_unary_evilly_mixed(self, request_iterator, context):
        assert _REQUEST == await context.read()
        request_count = 0
        async for request in request_iterator:
            assert _REQUEST == request
            request_count += 1
        assert _NUM_STREAM_REQUESTS - 1 == request_count
        return _RESPONSE

    async def _stream_stream_async_gen(self, request_iterator, unused_context):
        request_count = 0
        async for request in request_iterator:
            assert _REQUEST == request
            request_count += 1
        assert _NUM_STREAM_REQUESTS == request_count

        for _ in range(_NUM_STREAM_RESPONSES):
            yield _RESPONSE

    async def _stream_stream_reader_writer(self, unused_request, context):
        for _ in range(_NUM_STREAM_REQUESTS):
            assert _REQUEST == await context.read()
        for _ in range(_NUM_STREAM_RESPONSES):
            await context.write(_RESPONSE)

    async def _stream_stream_evilly_mixed(self, request_iterator, context):
        assert _REQUEST == await context.read()
        request_count = 0
        async for request in request_iterator:
            assert _REQUEST == request
            request_count += 1
        assert _NUM_STREAM_REQUESTS - 1 == request_count

        yield _RESPONSE
        for _ in range(_NUM_STREAM_RESPONSES - 1):
            await context.write(_RESPONSE)

    async def _error_in_stream_stream(self, request_iterator, unused_context):
        async for request in request_iterator:
            assert _REQUEST == request
            raise RuntimeError('A testing RuntimeError!')
        yield _RESPONSE

    async def _value_error_in_stream_unary(self, request_iterator, context):
        request_count = 0
        async for request in request_iterator:
            assert _REQUEST == request
            request_count += 1
            if request_count >= 1:
                raise ValueError('A testing RuntimeError!')

    async def _error_without_raise_in_unary_unary(self, request, context):
        assert _REQUEST == request
        context.set_code(grpc.StatusCode.INTERNAL)

    async def _error_without_raise_in_stream_stream(self, request_iterator,
                                                    context):
        async for request in request_iterator:
            assert _REQUEST == request
        context.set_code(grpc.StatusCode.INTERNAL)

    async def _invalid_trailing_metadata(self, request, context):
        assert _REQUEST == request
        for invalid_metadata in [
                42, {}, {
                    'error': 'error'
                }, [{
                    'error': "error"
                }]
        ]:
            try:
                context.set_trailing_metadata(invalid_metadata)
            except TypeError:
                pass
            else:
                raise ValueError(
                    f'No TypeError raised for invalid metadata: {invalid_metadata}'
                )

        await context.abort(grpc.StatusCode.DATA_LOSS,
                            details="invalid abort",
                            trailing_metadata=({
                                'error': ('error1', 'error2')
                            }))

    def service(self, handler_details):
        if not self._called.done():
            self._called.set_result(None)
        return self._routing_table.get(handler_details.method)

    async def wait_for_call(self):
        await self._called


async def _start_test_server():
    server = aio.server()
    port = server.add_insecure_port('[::]:0')
    generic_handler = _GenericHandler()
    server.add_generic_rpc_handlers((generic_handler,))
    await server.start()
    return 'localhost:%d' % port, server, generic_handler


class TestServer(AioTestBase):

    async def setUp(self):
        addr, self._server, self._generic_handler = await _start_test_server()
        self._channel = aio.insecure_channel(addr)

    async def tearDown(self):
        await self._channel.close()
        await self._server.stop(None)

    async def test_unary_unary(self):
        unary_unary_call = self._channel.unary_unary(_SIMPLE_UNARY_UNARY)
        response = await unary_unary_call(_REQUEST)
        self.assertEqual(response, _RESPONSE)

    async def test_unary_stream_async_generator(self):
        unary_stream_call = self._channel.unary_stream(_UNARY_STREAM_ASYNC_GEN)
        call = unary_stream_call(_REQUEST)

        response_cnt = 0
        async for response in call:
            response_cnt += 1
            self.assertEqual(_RESPONSE, response)

        self.assertEqual(_NUM_STREAM_RESPONSES, response_cnt)
        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_unary_stream_reader_writer(self):
        unary_stream_call = self._channel.unary_stream(
            _UNARY_STREAM_READER_WRITER)
        call = unary_stream_call(_REQUEST)

        for _ in range(_NUM_STREAM_RESPONSES):
            response = await call.read()
            self.assertEqual(_RESPONSE, response)

        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_unary_stream_evilly_mixed(self):
        unary_stream_call = self._channel.unary_stream(
            _UNARY_STREAM_EVILLY_MIXED)
        call = unary_stream_call(_REQUEST)

        # Uses reader API
        self.assertEqual(_RESPONSE, await call.read())

        # Uses async generator API, mixed!
        with self.assertRaises(aio.UsageError):
            async for response in call:
                self.assertEqual(_RESPONSE, response)

    async def test_stream_unary_async_generator(self):
        stream_unary_call = self._channel.stream_unary(_STREAM_UNARY_ASYNC_GEN)
        call = stream_unary_call()

        for _ in range(_NUM_STREAM_REQUESTS):
            await call.write(_REQUEST)
        await call.done_writing()

        response = await call
        self.assertEqual(_RESPONSE, response)
        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_stream_unary_async_generator_with_request_iter(self):
        stream_unary_call = self._channel.stream_unary(_STREAM_UNARY_ASYNC_GEN)

        finished = False

        def request_gen():
            for _ in range(_NUM_STREAM_REQUESTS):
                yield _REQUEST
            nonlocal finished
            finished = True

        call = stream_unary_call(request_gen())

        response = await call
        self.assertEqual(_RESPONSE, response)
        self.assertEqual(await call.code(), grpc.StatusCode.OK)
        self.assertEqual(finished, True)

    async def test_stream_unary_reader_writer(self):
        stream_unary_call = self._channel.stream_unary(
            _STREAM_UNARY_READER_WRITER)
        call = stream_unary_call()

        for _ in range(_NUM_STREAM_REQUESTS):
            await call.write(_REQUEST)
        await call.done_writing()

        response = await call
        self.assertEqual(_RESPONSE, response)
        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_stream_unary_evilly_mixed(self):
        stream_unary_call = self._channel.stream_unary(
            _STREAM_UNARY_EVILLY_MIXED)
        call = stream_unary_call()

        for _ in range(_NUM_STREAM_REQUESTS):
            await call.write(_REQUEST)
        await call.done_writing()

        response = await call
        self.assertEqual(_RESPONSE, response)
        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_stream_stream_async_generator(self):
        stream_stream_call = self._channel.stream_stream(
            _STREAM_STREAM_ASYNC_GEN)
        call = stream_stream_call()

        for _ in range(_NUM_STREAM_REQUESTS):
            await call.write(_REQUEST)
        await call.done_writing()

        for _ in range(_NUM_STREAM_RESPONSES):
            response = await call.read()
            self.assertEqual(_RESPONSE, response)

        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_stream_stream_reader_writer(self):
        stream_stream_call = self._channel.stream_stream(
            _STREAM_STREAM_READER_WRITER)
        call = stream_stream_call()

        for _ in range(_NUM_STREAM_REQUESTS):
            await call.write(_REQUEST)
        await call.done_writing()

        for _ in range(_NUM_STREAM_RESPONSES):
            response = await call.read()
            self.assertEqual(_RESPONSE, response)

        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_stream_stream_evilly_mixed(self):
        stream_stream_call = self._channel.stream_stream(
            _STREAM_STREAM_EVILLY_MIXED)
        call = stream_stream_call()

        for _ in range(_NUM_STREAM_REQUESTS):
            await call.write(_REQUEST)
        await call.done_writing()

        for _ in range(_NUM_STREAM_RESPONSES):
            response = await call.read()
            self.assertEqual(_RESPONSE, response)

        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_shutdown(self):
        await self._server.stop(None)
        # Ensures no SIGSEGV triggered, and ends within timeout.

    async def test_shutdown_after_call(self):
        await self._channel.unary_unary(_SIMPLE_UNARY_UNARY)(_REQUEST)

        await self._server.stop(None)

    async def test_graceful_shutdown_success(self):
        call = self._channel.unary_unary(_BLOCK_BRIEFLY)(_REQUEST)
        await self._generic_handler.wait_for_call()

        shutdown_start_time = time.time()
        await self._server.stop(test_constants.SHORT_TIMEOUT)
        grace_period_length = time.time() - shutdown_start_time
        self.assertGreater(grace_period_length,
                           test_constants.SHORT_TIMEOUT / 3)

        # Validates the states.
        self.assertEqual(_RESPONSE, await call)
        self.assertTrue(call.done())

    async def test_graceful_shutdown_failed(self):
        call = self._channel.unary_unary(_BLOCK_FOREVER)(_REQUEST)
        await self._generic_handler.wait_for_call()

        await self._server.stop(test_constants.SHORT_TIMEOUT)

        with self.assertRaises(aio.AioRpcError) as exception_context:
            await call
        self.assertEqual(grpc.StatusCode.UNAVAILABLE,
                         exception_context.exception.code())

    async def test_concurrent_graceful_shutdown(self):
        call = self._channel.unary_unary(_BLOCK_BRIEFLY)(_REQUEST)
        await self._generic_handler.wait_for_call()

        # Expects the shortest grace period to be effective.
        shutdown_start_time = time.time()
        await asyncio.gather(
            self._server.stop(test_constants.LONG_TIMEOUT),
            self._server.stop(test_constants.SHORT_TIMEOUT),
            self._server.stop(test_constants.LONG_TIMEOUT),
        )
        grace_period_length = time.time() - shutdown_start_time
        self.assertGreater(grace_period_length,
                           test_constants.SHORT_TIMEOUT / 3)

        self.assertEqual(_RESPONSE, await call)
        self.assertTrue(call.done())

    async def test_concurrent_graceful_shutdown_immediate(self):
        call = self._channel.unary_unary(_BLOCK_FOREVER)(_REQUEST)
        await self._generic_handler.wait_for_call()

        # Expects no grace period, due to the "server.stop(None)".
        await asyncio.gather(
            self._server.stop(test_constants.LONG_TIMEOUT),
            self._server.stop(None),
            self._server.stop(test_constants.SHORT_TIMEOUT),
            self._server.stop(test_constants.LONG_TIMEOUT),
        )

        with self.assertRaises(aio.AioRpcError) as exception_context:
            await call
        self.assertEqual(grpc.StatusCode.UNAVAILABLE,
                         exception_context.exception.code())

    async def test_shutdown_before_call(self):
        await self._server.stop(None)

        # Ensures the server is cleaned up at this point.
        # Some proper exception should be raised.
        with self.assertRaises(aio.AioRpcError):
            await self._channel.unary_unary(_SIMPLE_UNARY_UNARY)(_REQUEST)

    async def test_unimplemented(self):
        call = self._channel.unary_unary(_UNIMPLEMENTED_METHOD)
        with self.assertRaises(aio.AioRpcError) as exception_context:
            await call(_REQUEST)
        rpc_error = exception_context.exception
        self.assertEqual(grpc.StatusCode.UNIMPLEMENTED, rpc_error.code())

    async def test_shutdown_during_stream_stream(self):
        stream_stream_call = self._channel.stream_stream(
            _STREAM_STREAM_ASYNC_GEN)
        call = stream_stream_call()

        # Don't half close the RPC yet, keep it alive.
        await call.write(_REQUEST)
        await self._server.stop(None)

        self.assertEqual(grpc.StatusCode.UNAVAILABLE, await call.code())
        # No segfault

    async def test_error_in_stream_stream(self):
        stream_stream_call = self._channel.stream_stream(
            _ERROR_IN_STREAM_STREAM)
        call = stream_stream_call()

        # Don't half close the RPC yet, keep it alive.
        await call.write(_REQUEST)

        # Don't segfault here
        self.assertEqual(grpc.StatusCode.UNKNOWN, await call.code())

    async def test_error_without_raise_in_unary_unary(self):
        call = self._channel.unary_unary(_ERROR_WITHOUT_RAISE_IN_UNARY_UNARY)(
            _REQUEST)

        with self.assertRaises(aio.AioRpcError) as exception_context:
            await call

        rpc_error = exception_context.exception
        self.assertEqual(grpc.StatusCode.INTERNAL, rpc_error.code())

    async def test_error_without_raise_in_stream_stream(self):
        call = self._channel.stream_stream(
            _ERROR_WITHOUT_RAISE_IN_STREAM_STREAM)()

        for _ in range(_NUM_STREAM_REQUESTS):
            await call.write(_REQUEST)
        await call.done_writing()

        self.assertEqual(grpc.StatusCode.INTERNAL, await call.code())

    async def test_error_in_stream_unary(self):
        stream_unary_call = self._channel.stream_unary(_ERROR_IN_STREAM_UNARY)

        async def request_gen():
            for _ in range(_NUM_STREAM_REQUESTS):
                yield _REQUEST

        call = stream_unary_call(request_gen())

        with self.assertRaises(aio.AioRpcError) as exception_context:
            await call
        rpc_error = exception_context.exception
        self.assertEqual(grpc.StatusCode.UNKNOWN, rpc_error.code())

    async def test_port_binding_exception(self):
        server = aio.server(options=(('grpc.so_reuseport', 0),))
        port = server.add_insecure_port('localhost:0')
        bind_address = "localhost:%d" % port

        with self.assertRaises(RuntimeError):
            server.add_insecure_port(bind_address)

        server_credentials = grpc.ssl_server_credentials([
            (resources.private_key(), resources.certificate_chain())
        ])
        with self.assertRaises(RuntimeError):
            server.add_secure_port(bind_address, server_credentials)

    async def test_maximum_concurrent_rpcs(self):
        # Build the server with concurrent rpc argument
        server = aio.server(maximum_concurrent_rpcs=_MAXIMUM_CONCURRENT_RPCS)
        port = server.add_insecure_port('localhost:0')
        bind_address = "localhost:%d" % port
        server.add_generic_rpc_handlers((_GenericHandler(),))
        await server.start()
        # Build the channel
        channel = aio.insecure_channel(bind_address)
        # Deplete the concurrent quota with 3 times of max RPCs
        rpcs = []
        for _ in range(3 * _MAXIMUM_CONCURRENT_RPCS):
            rpcs.append(channel.unary_unary(_BLOCK_BRIEFLY)(_REQUEST))
        task = self.loop.create_task(
            asyncio.wait(rpcs, return_when=asyncio.FIRST_EXCEPTION))
        # Each batch took test_constants.SHORT_TIMEOUT /2
        start_time = time.time()
        await task
        elapsed_time = time.time() - start_time
        self.assertGreater(elapsed_time, test_constants.SHORT_TIMEOUT * 3 / 2)
        # Clean-up
        await channel.close()
        await server.stop(0)

    async def test_invalid_trailing_metadata(self):
        call = self._channel.unary_unary(_INVALID_TRAILING_METADATA)(_REQUEST)

        with self.assertRaises(aio.AioRpcError) as exception_context:
            await call

        rpc_error = exception_context.exception
        self.assertEqual(grpc.StatusCode.UNKNOWN, rpc_error.code())
        self.assertIn('trailing', rpc_error.details())


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    unittest.main(verbosity=2)



#!/usr/bin/env python3

# Copyright 2016 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import subprocess
import sys

import yaml

errors = 0

os.chdir(os.path.join(os.path.dirname(sys.argv[0]), '../../..'))

# hack import paths to pick up extra code
sys.path.insert(0, os.path.abspath('tools/buildgen/plugins'))
from expand_version import Version

try:
    branch_name = subprocess.check_output('git rev-parse --abbrev-ref HEAD',
                                          shell=True).decode()
except:
    print('WARNING: not a git repository')
    branch_name = None

if branch_name is not None:
    m = re.match(r'^release-([0-9]+)_([0-9]+)$', branch_name)
    if m:
        print('RELEASE branch')
        # version number should align with the branched version
        check_version = lambda version: (version.major == int(m.group(1)) and
                                         version.minor == int(m.group(2)))
        warning = 'Version key "%%s" value "%%s" should have a major version %s and minor version %s' % (
            m.group(1), m.group(2))
    elif re.match(r'^debian/.*$', branch_name):
        # no additional version checks for debian branches
        check_version = lambda version: True
    else:
        # all other branches should have a -dev tag
        check_version = lambda version: version.tag == 'dev'
        warning = 'Version key "%s" value "%s" should have a -dev tag'
else:
    check_version = lambda version: True

with open('build_handwritten.yaml', 'r') as f:
    build_yaml = yaml.load(f.read())

settings = build_yaml['settings']

top_version = Version(settings['version'])
if not check_version(top_version):
    errors += 1
    print((warning % ('version', top_version)))

for tag, value in list(settings.items()):
    if re.match(r'^[a-z]+_version$', tag):
        value = Version(value)
        if tag != 'core_version':
            if value.major != top_version.major:
                errors += 1
                print(('major version mismatch on %s: %d vs %d' %
                       (tag, value.major, top_version.major)))
            if value.minor != top_version.minor:
                errors += 1
                print(('minor version mismatch on %s: %d vs %d' %
                       (tag, value.minor, top_version.minor)))
        if not check_version(value):
            errors += 1
            print((warning % (tag, value)))

sys.exit(errors)



#!/usr/bin/env python3

# Copyright 2016 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Explicitly ban select functions from being used in src/core/**.

Most of these functions have internal versions that should be used instead."""

import os
import sys

os.chdir(os.path.join(os.path.dirname(sys.argv[0]), '../../..'))

# map of banned function signature to allowlist
BANNED_EXCEPT = {
    'grpc_slice_from_static_buffer(': ['src/core/lib/slice/slice.cc'],
    'grpc_resource_quota_ref(': ['src/core/lib/resource_quota/api.cc'],
    'grpc_resource_quota_unref(': [
        'src/core/lib/resource_quota/api.cc', 'src/core/lib/surface/server.cc'
    ],
    'grpc_slice_buffer_destroy(': ['src/core/lib/slice/slice_buffer_api.cc'],
    'grpc_slice_buffer_reset_and_unref(': [
        'src/core/lib/slice/slice_buffer_api.cc'
    ],
    'grpc_slice_ref(': ['src/core/lib/slice/slice_api.cc'],
    'grpc_slice_unref(': ['src/core/lib/slice/slice_api.cc'],
    'grpc_error_create(': [
        'src/core/lib/iomgr/error.cc', 'src/core/lib/iomgr/error_cfstream.cc'
    ],
    'grpc_error_ref(': ['src/core/lib/iomgr/error.cc'],
    'grpc_error_unref(': ['src/core/lib/iomgr/error.cc'],
    'grpc_os_error(': ['src/core/lib/iomgr/error.cc'],
    'grpc_wsa_error(': ['src/core/lib/iomgr/error.cc'],
    'grpc_log_if_error(': ['src/core/lib/iomgr/error.cc'],
    'grpc_slice_malloc(': ['src/core/lib/slice/slice.cc'],
    'grpc_call_cancel(': ['src/core/lib/surface/call.cc'],
    'grpc_closure_create(': ['src/core/lib/iomgr/closure.cc'],
    'grpc_closure_init(': ['src/core/lib/iomgr/closure.cc'],
    'grpc_closure_sched(': ['src/core/lib/iomgr/closure.cc'],
    'grpc_closure_run(': ['src/core/lib/iomgr/closure.cc'],
    'grpc_closure_list_sched(': ['src/core/lib/iomgr/closure.cc'],
    'grpc_error*': ['src/core/lib/iomgr/error.cc'],
    'grpc_error_string': ['src/core/lib/iomgr/error.cc'],
    # std::random_device needs /dev/random which is not available on all linuxes that we support.
    # Any usage must be optional and opt-in, so that those platforms can use gRPC without problem.
    'std::random_device': [
        'src/core/ext/filters/client_channel/lb_policy/rls/rls.cc',
        'src/core/ext/filters/client_channel/resolver/google_c2p/google_c2p_resolver.cc',
    ]
}

errors = 0
num_files = 0
for root, dirs, files in os.walk('src/core'):
    if root.startswith('src/core/tsi'):
        continue
    for filename in files:
        num_files += 1
        path = os.path.join(root, filename)
        if os.path.splitext(path)[1] != '.cc':
            continue
        with open(path) as f:
            text = f.read()
        for banned, exceptions in list(BANNED_EXCEPT.items()):
            if path in exceptions:
                continue
            if banned in text:
                print(('Illegal use of "%s" in %s' % (banned, path)))
                errors += 1

assert errors == 0
# This check comes about from this issue:
# https://github.com/grpc/grpc/issues/15381
# Basically, a change rendered this script useless and we did not realize it.
# This dumb check ensures that this type of issue doesn't occur again.
assert num_files > 300  # we definitely have more than 300 files



# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides distutils command classes for the gRPC Python setup process."""

from distutils import errors as _errors
import glob
import os
import os.path
import platform
import re
import shutil
import sys

import setuptools
from setuptools.command import build_ext
from setuptools.command import build_py
from setuptools.command import easy_install
from setuptools.command import install
from setuptools.command import test

PYTHON_STEM = os.path.dirname(os.path.abspath(__file__))
GRPC_STEM = os.path.abspath(PYTHON_STEM + '../../../../')
GRPC_PROTO_STEM = os.path.join(GRPC_STEM, 'src', 'proto')
PROTO_STEM = os.path.join(PYTHON_STEM, 'src', 'proto')
PYTHON_PROTO_TOP_LEVEL = os.path.join(PYTHON_STEM, 'src')


class CommandError(object):
    pass


class GatherProto(setuptools.Command):

    description = 'gather proto dependencies'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # TODO(atash) ensure that we're running from the repository directory when
        # this command is used
        try:
            shutil.rmtree(PROTO_STEM)
        except Exception as error:
            # We don't care if this command fails
            pass
        shutil.copytree(GRPC_PROTO_STEM, PROTO_STEM)
        for root, _, _ in os.walk(PYTHON_PROTO_TOP_LEVEL):
            path = os.path.join(root, '__init__.py')
            open(path, 'a').close()


class BuildPy(build_py.build_py):
    """Custom project build command."""

    def run(self):
        try:
            self.run_command('build_package_protos')
        except CommandError as error:
            sys.stderr.write('warning: %s\n' % error.message)
        build_py.build_py.run(self)


class TestLite(setuptools.Command):
    """Command to run tests without fetching or building anything."""

    description = 'run tests without fetching or building anything.'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        # distutils requires this override.
        pass

    def run(self):
        self._add_eggs_to_path()

        import tests
        loader = tests.Loader()
        loader.loadTestsFromNames(['tests'])
        runner = tests.Runner(dedicated_threads=True)
        result = runner.run(loader.suite)
        if not result.wasSuccessful():
            sys.exit('Test failure')

    def _add_eggs_to_path(self):
        """Fetch install and test requirements"""
        self.distribution.fetch_build_eggs(self.distribution.install_requires)
        self.distribution.fetch_build_eggs(self.distribution.tests_require)


class TestPy3Only(setuptools.Command):
    """Command to run tests for Python 3+ features.

    This does not include asyncio tests, which are housed in a separate
    directory.
    """

    description = 'run tests for py3+ features'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self._add_eggs_to_path()
        import tests
        loader = tests.Loader()
        loader.loadTestsFromNames(['tests_py3_only'])
        runner = tests.Runner()
        result = runner.run(loader.suite)
        if not result.wasSuccessful():
            sys.exit('Test failure')

    def _add_eggs_to_path(self):
        self.distribution.fetch_build_eggs(self.distribution.install_requires)
        self.distribution.fetch_build_eggs(self.distribution.tests_require)


class TestAio(setuptools.Command):
    """Command to run aio tests without fetching or building anything."""

    description = 'run aio tests without fetching or building anything.'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self._add_eggs_to_path()

        import tests
        loader = tests.Loader()
        loader.loadTestsFromNames(['tests_aio'])
        # Even without dedicated threads, the framework will somehow spawn a
        # new thread for tests to run upon. New thread doesn't have event loop
        # attached by default, so initialization is needed.
        runner = tests.Runner(dedicated_threads=False)
        result = runner.run(loader.suite)
        if not result.wasSuccessful():
            sys.exit('Test failure')

    def _add_eggs_to_path(self):
        """Fetch install and test requirements"""
        self.distribution.fetch_build_eggs(self.distribution.install_requires)
        self.distribution.fetch_build_eggs(self.distribution.tests_require)


class TestGevent(setuptools.Command):
    """Command to run tests w/gevent."""

    BANNED_TESTS = (
        # Fork support is not compatible with gevent
        'fork._fork_interop_test.ForkInteropTest',
        # These tests send a lot of RPCs and are really slow on gevent.  They will
        # eventually succeed, but need to dig into performance issues.
        'unit._cython._no_messages_server_completion_queue_per_call_test.Test.test_rpcs',
        'unit._cython._no_messages_single_server_completion_queue_test.Test.test_rpcs',
        'unit._compression_test',
        # TODO(https://github.com/grpc/grpc/issues/16890) enable this test
        'unit._cython._channel_test.ChannelTest.test_multiple_channels_lonely_connectivity',
        # I have no idea why this doesn't work in gevent, but it shouldn't even be
        # using the c-core
        'testing._client_test.ClientTest.test_infinite_request_stream_real_time',
        # TODO(https://github.com/grpc/grpc/issues/15743) enable this test
        'unit._session_cache_test.SSLSessionCacheTest.testSSLSessionCacheLRU',
        # TODO(https://github.com/grpc/grpc/issues/14789) enable this test
        'unit._server_ssl_cert_config_test',
        # TODO(https://github.com/grpc/grpc/issues/14901) enable this test
        'protoc_plugin._python_plugin_test.PythonPluginTest',
        'protoc_plugin._python_plugin_test.SimpleStubsPluginTest',
        # Beta API is unsupported for gevent
        'protoc_plugin.beta_python_plugin_test',
        'unit.beta._beta_features_test',
        # TODO(https://github.com/grpc/grpc/issues/15411) unpin gevent version
        # This test will stuck while running higher version of gevent
        'unit._auth_context_test.AuthContextTest.testSessionResumption',
        # TODO(https://github.com/grpc/grpc/issues/15411) enable these tests
        'unit._channel_ready_future_test.ChannelReadyFutureTest.test_immediately_connectable_channel_connectivity',
        "unit._cython._channel_test.ChannelTest.test_single_channel_lonely_connectivity",
        'unit._exit_test.ExitTest.test_in_flight_unary_unary_call',
        'unit._exit_test.ExitTest.test_in_flight_unary_stream_call',
        'unit._exit_test.ExitTest.test_in_flight_stream_unary_call',
        'unit._exit_test.ExitTest.test_in_flight_stream_stream_call',
        'unit._exit_test.ExitTest.test_in_flight_partial_unary_stream_call',
        'unit._exit_test.ExitTest.test_in_flight_partial_stream_unary_call',
        'unit._exit_test.ExitTest.test_in_flight_partial_stream_stream_call',
        # TODO(https://github.com/grpc/grpc/issues/18980): Reenable.
        'unit._signal_handling_test.SignalHandlingTest',
        'unit._metadata_flags_test',
        'health_check._health_servicer_test.HealthServicerTest.test_cancelled_watch_removed_from_watch_list',
        # TODO(https://github.com/grpc/grpc/issues/17330) enable these three tests
        'channelz._channelz_servicer_test.ChannelzServicerTest.test_many_subchannels',
        'channelz._channelz_servicer_test.ChannelzServicerTest.test_many_subchannels_and_sockets',
        'channelz._channelz_servicer_test.ChannelzServicerTest.test_streaming_rpc',
        # TODO(https://github.com/grpc/grpc/issues/15411) enable this test
        'unit._cython._channel_test.ChannelTest.test_negative_deadline_connectivity',
        # TODO(https://github.com/grpc/grpc/issues/15411) enable this test
        'unit._local_credentials_test.LocalCredentialsTest',
        # TODO(https://github.com/grpc/grpc/issues/22020) LocalCredentials
        # aren't supported with custom io managers.
        'unit._contextvars_propagation_test',
        'testing._time_test.StrictRealTimeTest',
    )
    BANNED_WINDOWS_TESTS = (
        # TODO(https://github.com/grpc/grpc/pull/15411) enable this test
        'unit._dns_resolver_test.DNSResolverTest.test_connect_loopback',
        # TODO(https://github.com/grpc/grpc/pull/15411) enable this test
        'unit._server_test.ServerTest.test_failed_port_binding_exception',
    )
    BANNED_MACOS_TESTS = (
        # TODO(https://github.com/grpc/grpc/issues/15411) enable this test
        'unit._dynamic_stubs_test.DynamicStubTest',)
    description = 'run tests with gevent.  Assumes grpc/gevent are installed'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        # distutils requires this override.
        pass

    def run(self):
        import gevent
        from gevent import monkey
        monkey.patch_all()

        threadpool = gevent.hub.get_hub().threadpool

        # Currently, each channel corresponds to a single native thread in the
        # gevent threadpool. Thus, when the unit test suite spins up hundreds of
        # channels concurrently, some will be starved out, causing the test to
        # increase in duration. We increase the max size here so this does not
        # happen.
        threadpool.maxsize = 1024
        threadpool.size = 32

        import grpc.experimental.gevent

        import tests
        grpc.experimental.gevent.init_gevent()

        import gevent

        import tests
        loader = tests.Loader()
        loader.loadTestsFromNames(['tests', 'tests_gevent'])
        runner = tests.Runner()
        if sys.platform == 'win32':
            runner.skip_tests(self.BANNED_TESTS + self.BANNED_WINDOWS_TESTS)
        elif sys.platform == 'darwin':
            runner.skip_tests(self.BANNED_TESTS + self.BANNED_MACOS_TESTS)
        else:
            runner.skip_tests(self.BANNED_TESTS)
        result = gevent.spawn(runner.run, loader.suite)
        result.join()
        if not result.value.wasSuccessful():
            sys.exit('Test failure')


class RunInterop(test.test):

    description = 'run interop test client/server'
    user_options = [
        ('args=', None, 'pass-thru arguments for the client/server'),
        ('client', None, 'flag indicating to run the client'),
        ('server', None, 'flag indicating to run the server'),
        ('use-asyncio', None, 'flag indicating to run the asyncio stack')
    ]

    def initialize_options(self):
        self.args = ''
        self.client = False
        self.server = False
        self.use_asyncio = False

    def finalize_options(self):
        if self.client and self.server:
            raise _errors.DistutilsOptionError(
                'you may only specify one of client or server')

    def run(self):
        if self.distribution.install_requires:
            self.distribution.fetch_build_eggs(
                self.distribution.install_requires)
        if self.distribution.tests_require:
            self.distribution.fetch_build_eggs(self.distribution.tests_require)
        if self.client:
            self.run_client()
        elif self.server:
            self.run_server()

    def run_server(self):
        # We import here to ensure that our setuptools parent has had a chance to
        # edit the Python system path.
        if self.use_asyncio:
            import asyncio

            from tests_aio.interop import server
            sys.argv[1:] = self.args.split()
            asyncio.get_event_loop().run_until_complete(server.serve())
        else:
            from tests.interop import server
            sys.argv[1:] = self.args.split()
            server.serve()

    def run_client(self):
        # We import here to ensure that our setuptools parent has had a chance to
        # edit the Python system path.
        from tests.interop import client
        sys.argv[1:] = self.args.split()
        client.test_interoperability()


class RunFork(test.test):

    description = 'run fork test client'
    user_options = [('args=', 'a', 'pass-thru arguments for the client')]

    def initialize_options(self):
        self.args = ''

    def finalize_options(self):
        # distutils requires this override.
        pass

    def run(self):
        if self.distribution.install_requires:
            self.distribution.fetch_build_eggs(
                self.distribution.install_requires)
        if self.distribution.tests_require:
            self.distribution.fetch_build_eggs(self.distribution.tests_require)
        # We import here to ensure that our setuptools parent has had a chance to
        # edit the Python system path.
        from tests.fork import client
        sys.argv[1:] = self.args.split()
        client.test_fork()



# Copyright 2021 The gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests xDS server and channel credentials."""

from concurrent import futures
import contextlib
import logging
import unittest

import grpc
import grpc.experimental

from tests.unit import resources
from tests.unit import test_common


class _GenericHandler(grpc.GenericRpcHandler):

    def service(self, handler_call_details):
        return grpc.unary_unary_rpc_method_handler(
            lambda request, unused_context: request)


@contextlib.contextmanager
def xds_channel_server_without_xds(server_fallback_creds):
    server = grpc.server(futures.ThreadPoolExecutor())
    server.add_generic_rpc_handlers((_GenericHandler(),))
    server_server_fallback_creds = grpc.ssl_server_credentials(
        ((resources.private_key(), resources.certificate_chain()),))
    server_creds = grpc.xds_server_credentials(server_fallback_creds)
    port = server.add_secure_port("localhost:0", server_creds)
    server.start()
    try:
        yield "localhost:{}".format(port)
    finally:
        server.stop(None)


class XdsCredentialsTest(unittest.TestCase):

    def test_xds_creds_fallback_ssl(self):
        # Since there is no xDS server, the fallback credentials will be used.
        # In this case, SSL credentials.
        server_fallback_creds = grpc.ssl_server_credentials(
            ((resources.private_key(), resources.certificate_chain()),))
        with xds_channel_server_without_xds(
                server_fallback_creds) as server_address:
            override_options = (("grpc.ssl_target_name_override",
                                 "foo.test.google.fr"),)
            channel_fallback_creds = grpc.ssl_channel_credentials(
                root_certificates=resources.test_root_certificates(),
                private_key=resources.private_key(),
                certificate_chain=resources.certificate_chain())
            channel_creds = grpc.xds_channel_credentials(channel_fallback_creds)
            with grpc.secure_channel(server_address,
                                     channel_creds,
                                     options=override_options) as channel:
                request = b"abc"
                response = channel.unary_unary("/test/method")(
                    request, wait_for_ready=True)
                self.assertEqual(response, request)

    def test_xds_creds_fallback_insecure(self):
        # Since there is no xDS server, the fallback credentials will be used.
        # In this case, insecure.
        server_fallback_creds = grpc.insecure_server_credentials()
        with xds_channel_server_without_xds(
                server_fallback_creds) as server_address:
            channel_fallback_creds = grpc.experimental.insecure_channel_credentials(
            )
            channel_creds = grpc.xds_channel_credentials(channel_fallback_creds)
            with grpc.secure_channel(server_address, channel_creds) as channel:
                request = b"abc"
                response = channel.unary_unary("/test/method")(
                    request, wait_for_ready=True)
                self.assertEqual(response, request)

    def test_start_xds_server(self):
        server = grpc.server(futures.ThreadPoolExecutor(), xds=True)
        server.add_generic_rpc_handlers((_GenericHandler(),))
        server_fallback_creds = grpc.insecure_server_credentials()
        server_creds = grpc.xds_server_credentials(server_fallback_creds)
        port = server.add_secure_port("localhost:0", server_creds)
        server.start()
        server.stop(None)
        # No exceptions thrown. A more comprehensive suite of tests will be
        # provided by the interop tests.


if __name__ == "__main__":
    logging.basicConfig()
    unittest.main()



# Copyright 2020 The gRPC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Python example of utilizing wait-for-ready flag."""

import asyncio
from contextlib import contextmanager
import logging
import socket
from typing import Iterable

import grpc

helloworld_pb2, helloworld_pb2_grpc = grpc.protos_and_services(
    "helloworld.proto")

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)


@contextmanager
def get_free_loopback_tcp_port() -> Iterable[str]:
    if socket.has_ipv6:
        tcp_socket = socket.socket(socket.AF_INET6)
    else:
        tcp_socket = socket.socket(socket.AF_INET)
    tcp_socket.bind(('', 0))
    address_tuple = tcp_socket.getsockname()
    yield f"localhost:{address_tuple[1]}"
    tcp_socket.close()


class Greeter(helloworld_pb2_grpc.GreeterServicer):

    async def SayHello(self, request: helloworld_pb2.HelloRequest,
                       unused_context) -> helloworld_pb2.HelloReply:
        return helloworld_pb2.HelloReply(message=f'Hello, {request.name}!')


def create_server(server_address: str):
    server = grpc.aio.server()
    helloworld_pb2_grpc.add_GreeterServicer_to_server(Greeter(), server)
    bound_port = server.add_insecure_port(server_address)
    assert bound_port == int(server_address.split(':')[-1])
    return server


async def process(stub: helloworld_pb2_grpc.GreeterStub,
                  wait_for_ready: bool = None) -> None:
    try:
        response = await stub.SayHello(helloworld_pb2.HelloRequest(name='you'),
                                       wait_for_ready=wait_for_ready)
        message = response.message
    except grpc.aio.AioRpcError as rpc_error:
        assert rpc_error.code() == grpc.StatusCode.UNAVAILABLE
        assert not wait_for_ready
        message = rpc_error
    else:
        assert wait_for_ready
    _LOGGER.info("Wait-for-ready %s, client received: %s",
                 "enabled" if wait_for_ready else "disabled", message)


async def main() -> None:
    # Pick a random free port
    with get_free_loopback_tcp_port() as server_address:
        # Create gRPC channel
        channel = grpc.aio.insecure_channel(server_address)
        stub = helloworld_pb2_grpc.GreeterStub(channel)

        # Fire an RPC without wait_for_ready
        fail_fast_task = asyncio.get_event_loop().create_task(
            process(stub, wait_for_ready=False))
        # Fire an RPC with wait_for_ready
        wait_for_ready_task = asyncio.get_event_loop().create_task(
            process(stub, wait_for_ready=True))

    # Wait for the channel entering TRANSIENT FAILURE state.
    state = channel.get_state()
    while state != grpc.ChannelConnectivity.TRANSIENT_FAILURE:
        await channel.wait_for_state_change(state)
        state = channel.get_state()

    # Start the server to handle the RPC
    server = create_server(server_address)
    await server.start()

    # Expected to fail with StatusCode.UNAVAILABLE.
    await fail_fast_task
    # Expected to success.
    await wait_for_ready_task

    await server.stop(None)
    await channel.close()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    asyncio.get_event_loop().run_until_complete(main())



# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Constants and functions for data used in interoperability testing."""

import argparse
import os
import pkgutil

_ROOT_CERTIFICATES_RESOURCE_PATH = 'credentials/ca.pem'
_PRIVATE_KEY_RESOURCE_PATH = 'credentials/server1.key'
_CERTIFICATE_CHAIN_RESOURCE_PATH = 'credentials/server1.pem'


def test_root_certificates():
    return pkgutil.get_data(__name__, _ROOT_CERTIFICATES_RESOURCE_PATH)


def private_key():
    return pkgutil.get_data(__name__, _PRIVATE_KEY_RESOURCE_PATH)


def certificate_chain():
    return pkgutil.get_data(__name__, _CERTIFICATE_CHAIN_RESOURCE_PATH)


def parse_bool(value):
    if value == 'true':
        return True
    if value == 'false':
        return False
    raise argparse.ArgumentTypeError('Only true/false allowed')



# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test of gRPC Python interceptors."""

import collections
from concurrent import futures
import itertools
import logging
import os
import threading
import unittest

import grpc
from grpc.framework.foundation import logging_pool

from tests.unit import test_common
from tests.unit.framework.common import test_constants
from tests.unit.framework.common import test_control

_SERIALIZE_REQUEST = lambda bytestring: bytestring * 2
_DESERIALIZE_REQUEST = lambda bytestring: bytestring[len(bytestring) // 2:]
_SERIALIZE_RESPONSE = lambda bytestring: bytestring * 3
_DESERIALIZE_RESPONSE = lambda bytestring: bytestring[:len(bytestring) // 3]

_EXCEPTION_REQUEST = b'\x09\x0a'

_UNARY_UNARY = '/test/UnaryUnary'
_UNARY_STREAM = '/test/UnaryStream'
_STREAM_UNARY = '/test/StreamUnary'
_STREAM_STREAM = '/test/StreamStream'


class _ApplicationErrorStandin(Exception):
    pass


class _Callback(object):

    def __init__(self):
        self._condition = threading.Condition()
        self._value = None
        self._called = False

    def __call__(self, value):
        with self._condition:
            self._value = value
            self._called = True
            self._condition.notify_all()

    def value(self):
        with self._condition:
            while not self._called:
                self._condition.wait()
            return self._value


class _Handler(object):

    def __init__(self, control):
        self._control = control

    def handle_unary_unary(self, request, servicer_context):
        self._control.control()
        if servicer_context is not None:
            servicer_context.set_trailing_metadata(((
                'testkey',
                'testvalue',
            ),))
        if request == _EXCEPTION_REQUEST:
            raise _ApplicationErrorStandin()
        return request

    def handle_unary_stream(self, request, servicer_context):
        if request == _EXCEPTION_REQUEST:
            raise _ApplicationErrorStandin()
        for _ in range(test_constants.STREAM_LENGTH):
            self._control.control()
            yield request
        self._control.control()
        if servicer_context is not None:
            servicer_context.set_trailing_metadata(((
                'testkey',
                'testvalue',
            ),))

    def handle_stream_unary(self, request_iterator, servicer_context):
        if servicer_context is not None:
            servicer_context.invocation_metadata()
        self._control.control()
        response_elements = []
        for request in request_iterator:
            self._control.control()
            response_elements.append(request)
        self._control.control()
        if servicer_context is not None:
            servicer_context.set_trailing_metadata(((
                'testkey',
                'testvalue',
            ),))
        if _EXCEPTION_REQUEST in response_elements:
            raise _ApplicationErrorStandin()
        return b''.join(response_elements)

    def handle_stream_stream(self, request_iterator, servicer_context):
        self._control.control()
        if servicer_context is not None:
            servicer_context.set_trailing_metadata(((
                'testkey',
                'testvalue',
            ),))
        for request in request_iterator:
            if request == _EXCEPTION_REQUEST:
                raise _ApplicationErrorStandin()
            self._control.control()
            yield request
        self._control.control()


class _MethodHandler(grpc.RpcMethodHandler):

    def __init__(self, request_streaming, response_streaming,
                 request_deserializer, response_serializer, unary_unary,
                 unary_stream, stream_unary, stream_stream):
        self.request_streaming = request_streaming
        self.response_streaming = response_streaming
        self.request_deserializer = request_deserializer
        self.response_serializer = response_serializer
        self.unary_unary = unary_unary
        self.unary_stream = unary_stream
        self.stream_unary = stream_unary
        self.stream_stream = stream_stream


class _GenericHandler(grpc.GenericRpcHandler):

    def __init__(self, handler):
        self._handler = handler

    def service(self, handler_call_details):
        if handler_call_details.method == _UNARY_UNARY:
            return _MethodHandler(False, False, None, None,
                                  self._handler.handle_unary_unary, None, None,
                                  None)
        elif handler_call_details.method == _UNARY_STREAM:
            return _MethodHandler(False, True, _DESERIALIZE_REQUEST,
                                  _SERIALIZE_RESPONSE, None,
                                  self._handler.handle_unary_stream, None, None)
        elif handler_call_details.method == _STREAM_UNARY:
            return _MethodHandler(True, False, _DESERIALIZE_REQUEST,
                                  _SERIALIZE_RESPONSE, None, None,
                                  self._handler.handle_stream_unary, None)
        elif handler_call_details.method == _STREAM_STREAM:
            return _MethodHandler(True, True, None, None, None, None, None,
                                  self._handler.handle_stream_stream)
        else:
            return None


def _unary_unary_multi_callable(channel):
    return channel.unary_unary(_UNARY_UNARY)


def _unary_stream_multi_callable(channel):
    return channel.unary_stream(_UNARY_STREAM,
                                request_serializer=_SERIALIZE_REQUEST,
                                response_deserializer=_DESERIALIZE_RESPONSE)


def _stream_unary_multi_callable(channel):
    return channel.stream_unary(_STREAM_UNARY,
                                request_serializer=_SERIALIZE_REQUEST,
                                response_deserializer=_DESERIALIZE_RESPONSE)


def _stream_stream_multi_callable(channel):
    return channel.stream_stream(_STREAM_STREAM)


class _ClientCallDetails(
        collections.namedtuple(
            '_ClientCallDetails',
            ('method', 'timeout', 'metadata', 'credentials')),
        grpc.ClientCallDetails):
    pass


class _GenericClientInterceptor(grpc.UnaryUnaryClientInterceptor,
                                grpc.UnaryStreamClientInterceptor,
                                grpc.StreamUnaryClientInterceptor,
                                grpc.StreamStreamClientInterceptor):

    def __init__(self, interceptor_function):
        self._fn = interceptor_function

    def intercept_unary_unary(self, continuation, client_call_details, request):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, False)
        response = continuation(new_details, next(new_request_iterator))
        return postprocess(response) if postprocess else response

    def intercept_unary_stream(self, continuation, client_call_details,
                               request):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, True)
        response_it = continuation(new_details, new_request_iterator)
        return postprocess(response_it) if postprocess else response_it

    def intercept_stream_unary(self, continuation, client_call_details,
                               request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator, True, False)
        response = continuation(new_details, next(new_request_iterator))
        return postprocess(response) if postprocess else response

    def intercept_stream_stream(self, continuation, client_call_details,
                                request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator, True, True)
        response_it = continuation(new_details, new_request_iterator)
        return postprocess(response_it) if postprocess else response_it


class _LoggingInterceptor(grpc.ServerInterceptor,
                          grpc.UnaryUnaryClientInterceptor,
                          grpc.UnaryStreamClientInterceptor,
                          grpc.StreamUnaryClientInterceptor,
                          grpc.StreamStreamClientInterceptor):

    def __init__(self, tag, record):
        self.tag = tag
        self.record = record

    def intercept_service(self, continuation, handler_call_details):
        self.record.append(self.tag + ':intercept_service')
        return continuation(handler_call_details)

    def intercept_unary_unary(self, continuation, client_call_details, request):
        self.record.append(self.tag + ':intercept_unary_unary')
        result = continuation(client_call_details, request)
        assert isinstance(
            result,
            grpc.Call), '{} ({}) is not an instance of grpc.Call'.format(
                result, type(result))
        assert isinstance(
            result,
            grpc.Future), '{} ({}) is not an instance of grpc.Future'.format(
                result, type(result))
        return result

    def intercept_unary_stream(self, continuation, client_call_details,
                               request):
        self.record.append(self.tag + ':intercept_unary_stream')
        return continuation(client_call_details, request)

    def intercept_stream_unary(self, continuation, client_call_details,
                               request_iterator):
        self.record.append(self.tag + ':intercept_stream_unary')
        result = continuation(client_call_details, request_iterator)
        assert isinstance(
            result,
            grpc.Call), '{} is not an instance of grpc.Call'.format(result)
        assert isinstance(
            result,
            grpc.Future), '{} is not an instance of grpc.Future'.format(result)
        return result

    def intercept_stream_stream(self, continuation, client_call_details,
                                request_iterator):
        self.record.append(self.tag + ':intercept_stream_stream')
        return continuation(client_call_details, request_iterator)


class _DefectiveClientInterceptor(grpc.UnaryUnaryClientInterceptor):

    def intercept_unary_unary(self, ignored_continuation,
                              ignored_client_call_details, ignored_request):
        raise test_control.Defect()


def _wrap_request_iterator_stream_interceptor(wrapper):

    def intercept_call(client_call_details, request_iterator, request_streaming,
                       ignored_response_streaming):
        if request_streaming:
            return client_call_details, wrapper(request_iterator), None
        else:
            return client_call_details, request_iterator, None

    return _GenericClientInterceptor(intercept_call)


def _append_request_header_interceptor(header, value):

    def intercept_call(client_call_details, request_iterator,
                       ignored_request_streaming, ignored_response_streaming):
        metadata = []
        if client_call_details.metadata:
            metadata = list(client_call_details.metadata)
        metadata.append((
            header,
            value,
        ))
        client_call_details = _ClientCallDetails(
            client_call_details.method, client_call_details.timeout, metadata,
            client_call_details.credentials)
        return client_call_details, request_iterator, None

    return _GenericClientInterceptor(intercept_call)


class _GenericServerInterceptor(grpc.ServerInterceptor):

    def __init__(self, fn):
        self._fn = fn

    def intercept_service(self, continuation, handler_call_details):
        return self._fn(continuation, handler_call_details)


def _filter_server_interceptor(condition, interceptor):

    def intercept_service(continuation, handler_call_details):
        if condition(handler_call_details):
            return interceptor.intercept_service(continuation,
                                                 handler_call_details)
        return continuation(handler_call_details)

    return _GenericServerInterceptor(intercept_service)


class InterceptorTest(unittest.TestCase):

    def setUp(self):
        self._control = test_control.PauseFailControl()
        self._handler = _Handler(self._control)
        self._server_pool = logging_pool.pool(test_constants.THREAD_CONCURRENCY)

        self._record = []
        conditional_interceptor = _filter_server_interceptor(
            lambda x: ('secret', '42') in x.invocation_metadata,
            _LoggingInterceptor('s3', self._record))

        self._server = grpc.server(self._server_pool,
                                   options=(('grpc.so_reuseport', 0),),
                                   interceptors=(
                                       _LoggingInterceptor('s1', self._record),
                                       conditional_interceptor,
                                       _LoggingInterceptor('s2', self._record),
                                   ))
        port = self._server.add_insecure_port('[::]:0')
        self._server.add_generic_rpc_handlers((_GenericHandler(self._handler),))
        self._server.start()

        self._channel = grpc.insecure_channel('localhost:%d' % port)

    def tearDown(self):
        self._server.stop(None)
        self._server_pool.shutdown(wait=True)
        self._channel.close()

    def testTripleRequestMessagesClientInterceptor(self):

        def triple(request_iterator):
            while True:
                try:
                    item = next(request_iterator)
                    yield item
                    yield item
                    yield item
                except StopIteration:
                    break

        interceptor = _wrap_request_iterator_stream_interceptor(triple)
        channel = grpc.intercept_channel(self._channel, interceptor)
        requests = tuple(
            b'\x07\x08' for _ in range(test_constants.STREAM_LENGTH))

        multi_callable = _stream_stream_multi_callable(channel)
        response_iterator = multi_callable(
            iter(requests),
            metadata=(
                ('test',
                 'InterceptedStreamRequestBlockingUnaryResponseWithCall'),))

        responses = tuple(response_iterator)
        self.assertEqual(len(responses), 3 * test_constants.STREAM_LENGTH)

        multi_callable = _stream_stream_multi_callable(self._channel)
        response_iterator = multi_callable(
            iter(requests),
            metadata=(
                ('test',
                 'InterceptedStreamRequestBlockingUnaryResponseWithCall'),))

        responses = tuple(response_iterator)
        self.assertEqual(len(responses), test_constants.STREAM_LENGTH)

    def testDefectiveClientInterceptor(self):
        interceptor = _DefectiveClientInterceptor()
        defective_channel = grpc.intercept_channel(self._channel, interceptor)

        request = b'\x07\x08'

        multi_callable = _unary_unary_multi_callable(defective_channel)
        call_future = multi_callable.future(
            request,
            metadata=(('test',
                       'InterceptedUnaryRequestBlockingUnaryResponse'),))

        self.assertIsNotNone(call_future.exception())
        self.assertEqual(call_future.code(), grpc.StatusCode.INTERNAL)

    def testInterceptedHeaderManipulationWithServerSideVerification(self):
        request = b'\x07\x08'

        channel = grpc.intercept_channel(
            self._channel, _append_request_header_interceptor('secret', '42'))
        channel = grpc.intercept_channel(
            channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        self._record[:] = []

        multi_callable = _unary_unary_multi_callable(channel)
        multi_callable.with_call(
            request,
            metadata=(
                ('test',
                 'InterceptedUnaryRequestBlockingUnaryResponseWithCall'),))

        self.assertSequenceEqual(self._record, [
            'c1:intercept_unary_unary', 'c2:intercept_unary_unary',
            's1:intercept_service', 's3:intercept_service',
            's2:intercept_service'
        ])

    def testInterceptedUnaryRequestBlockingUnaryResponse(self):
        request = b'\x07\x08'

        self._record[:] = []

        channel = grpc.intercept_channel(
            self._channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        multi_callable = _unary_unary_multi_callable(channel)
        multi_callable(
            request,
            metadata=(('test',
                       'InterceptedUnaryRequestBlockingUnaryResponse'),))

        self.assertSequenceEqual(self._record, [
            'c1:intercept_unary_unary', 'c2:intercept_unary_unary',
            's1:intercept_service', 's2:intercept_service'
        ])

    def testInterceptedUnaryRequestBlockingUnaryResponseWithError(self):
        request = _EXCEPTION_REQUEST

        self._record[:] = []

        channel = grpc.intercept_channel(
            self._channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        multi_callable = _unary_unary_multi_callable(channel)
        with self.assertRaises(grpc.RpcError) as exception_context:
            multi_callable(
                request,
                metadata=(('test',
                           'InterceptedUnaryRequestBlockingUnaryResponse'),))
        exception = exception_context.exception
        self.assertFalse(exception.cancelled())
        self.assertFalse(exception.running())
        self.assertTrue(exception.done())
        with self.assertRaises(grpc.RpcError):
            exception.result()
        self.assertIsInstance(exception.exception(), grpc.RpcError)

    def testInterceptedUnaryRequestBlockingUnaryResponseWithCall(self):
        request = b'\x07\x08'

        channel = grpc.intercept_channel(
            self._channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        self._record[:] = []

        multi_callable = _unary_unary_multi_callable(channel)
        multi_callable.with_call(
            request,
            metadata=(
                ('test',
                 'InterceptedUnaryRequestBlockingUnaryResponseWithCall'),))

        self.assertSequenceEqual(self._record, [
            'c1:intercept_unary_unary', 'c2:intercept_unary_unary',
            's1:intercept_service', 's2:intercept_service'
        ])

    def testInterceptedUnaryRequestFutureUnaryResponse(self):
        request = b'\x07\x08'

        self._record[:] = []
        channel = grpc.intercept_channel(
            self._channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        multi_callable = _unary_unary_multi_callable(channel)
        response_future = multi_callable.future(
            request,
            metadata=(('test', 'InterceptedUnaryRequestFutureUnaryResponse'),))
        response_future.result()

        self.assertSequenceEqual(self._record, [
            'c1:intercept_unary_unary', 'c2:intercept_unary_unary',
            's1:intercept_service', 's2:intercept_service'
        ])

    def testInterceptedUnaryRequestStreamResponse(self):
        request = b'\x37\x58'

        self._record[:] = []
        channel = grpc.intercept_channel(
            self._channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        multi_callable = _unary_stream_multi_callable(channel)
        response_iterator = multi_callable(
            request,
            metadata=(('test', 'InterceptedUnaryRequestStreamResponse'),))
        tuple(response_iterator)

        self.assertSequenceEqual(self._record, [
            'c1:intercept_unary_stream', 'c2:intercept_unary_stream',
            's1:intercept_service', 's2:intercept_service'
        ])

    def testInterceptedUnaryRequestStreamResponseWithError(self):
        request = _EXCEPTION_REQUEST

        self._record[:] = []
        channel = grpc.intercept_channel(
            self._channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        multi_callable = _unary_stream_multi_callable(channel)
        response_iterator = multi_callable(
            request,
            metadata=(('test', 'InterceptedUnaryRequestStreamResponse'),))
        with self.assertRaises(grpc.RpcError) as exception_context:
            tuple(response_iterator)
        exception = exception_context.exception
        self.assertFalse(exception.cancelled())
        self.assertFalse(exception.running())
        self.assertTrue(exception.done())
        with self.assertRaises(grpc.RpcError):
            exception.result()
        self.assertIsInstance(exception.exception(), grpc.RpcError)

    def testInterceptedStreamRequestBlockingUnaryResponse(self):
        requests = tuple(
            b'\x07\x08' for _ in range(test_constants.STREAM_LENGTH))
        request_iterator = iter(requests)

        self._record[:] = []
        channel = grpc.intercept_channel(
            self._channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        multi_callable = _stream_unary_multi_callable(channel)
        multi_callable(
            request_iterator,
            metadata=(('test',
                       'InterceptedStreamRequestBlockingUnaryResponse'),))

        self.assertSequenceEqual(self._record, [
            'c1:intercept_stream_unary', 'c2:intercept_stream_unary',
            's1:intercept_service', 's2:intercept_service'
        ])

    def testInterceptedStreamRequestBlockingUnaryResponseWithCall(self):
        requests = tuple(
            b'\x07\x08' for _ in range(test_constants.STREAM_LENGTH))
        request_iterator = iter(requests)

        self._record[:] = []
        channel = grpc.intercept_channel(
            self._channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        multi_callable = _stream_unary_multi_callable(channel)
        multi_callable.with_call(
            request_iterator,
            metadata=(
                ('test',
                 'InterceptedStreamRequestBlockingUnaryResponseWithCall'),))

        self.assertSequenceEqual(self._record, [
            'c1:intercept_stream_unary', 'c2:intercept_stream_unary',
            's1:intercept_service', 's2:intercept_service'
        ])

    def testInterceptedStreamRequestFutureUnaryResponse(self):
        requests = tuple(
            b'\x07\x08' for _ in range(test_constants.STREAM_LENGTH))
        request_iterator = iter(requests)

        self._record[:] = []
        channel = grpc.intercept_channel(
            self._channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        multi_callable = _stream_unary_multi_callable(channel)
        response_future = multi_callable.future(
            request_iterator,
            metadata=(('test', 'InterceptedStreamRequestFutureUnaryResponse'),))
        response_future.result()

        self.assertSequenceEqual(self._record, [
            'c1:intercept_stream_unary', 'c2:intercept_stream_unary',
            's1:intercept_service', 's2:intercept_service'
        ])

    def testInterceptedStreamRequestFutureUnaryResponseWithError(self):
        requests = tuple(
            _EXCEPTION_REQUEST for _ in range(test_constants.STREAM_LENGTH))
        request_iterator = iter(requests)

        self._record[:] = []
        channel = grpc.intercept_channel(
            self._channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        multi_callable = _stream_unary_multi_callable(channel)
        response_future = multi_callable.future(
            request_iterator,
            metadata=(('test', 'InterceptedStreamRequestFutureUnaryResponse'),))
        with self.assertRaises(grpc.RpcError) as exception_context:
            response_future.result()
        exception = exception_context.exception
        self.assertFalse(exception.cancelled())
        self.assertFalse(exception.running())
        self.assertTrue(exception.done())
        with self.assertRaises(grpc.RpcError):
            exception.result()
        self.assertIsInstance(exception.exception(), grpc.RpcError)

    def testInterceptedStreamRequestStreamResponse(self):
        requests = tuple(
            b'\x77\x58' for _ in range(test_constants.STREAM_LENGTH))
        request_iterator = iter(requests)

        self._record[:] = []
        channel = grpc.intercept_channel(
            self._channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        multi_callable = _stream_stream_multi_callable(channel)
        response_iterator = multi_callable(
            request_iterator,
            metadata=(('test', 'InterceptedStreamRequestStreamResponse'),))
        tuple(response_iterator)

        self.assertSequenceEqual(self._record, [
            'c1:intercept_stream_stream', 'c2:intercept_stream_stream',
            's1:intercept_service', 's2:intercept_service'
        ])

    def testInterceptedStreamRequestStreamResponseWithError(self):
        requests = tuple(
            _EXCEPTION_REQUEST for _ in range(test_constants.STREAM_LENGTH))
        request_iterator = iter(requests)

        self._record[:] = []
        channel = grpc.intercept_channel(
            self._channel, _LoggingInterceptor('c1', self._record),
            _LoggingInterceptor('c2', self._record))

        multi_callable = _stream_stream_multi_callable(channel)
        response_iterator = multi_callable(
            request_iterator,
            metadata=(('test', 'InterceptedStreamRequestStreamResponse'),))
        with self.assertRaises(grpc.RpcError) as exception_context:
            tuple(response_iterator)
        exception = exception_context.exception
        self.assertFalse(exception.cancelled())
        self.assertFalse(exception.running())
        self.assertTrue(exception.done())
        with self.assertRaises(grpc.RpcError):
            exception.result()
        self.assertIsInstance(exception.exception(), grpc.RpcError)


if __name__ == '__main__':
    logging.basicConfig()
    unittest.main(verbosity=2)



# Copyright 2016 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pkgutil
import unittest

import six

import tests


class SanityTest(unittest.TestCase):

    maxDiff = 32768

    TEST_PKG_MODULE_NAME = 'tests'
    TEST_PKG_PATH = 'tests'

    def testTestsJsonUpToDate(self):
        """Autodiscovers all test suites and checks that tests.json is up to date"""
        loader = tests.Loader()
        loader.loadTestsFromNames([self.TEST_PKG_MODULE_NAME])
        test_suite_names = sorted({
            test_case_class.id().rsplit('.', 1)[0] for test_case_class in
            tests._loader.iterate_suite_cases(loader.suite)
        })

        tests_json_string = pkgutil.get_data(self.TEST_PKG_PATH, 'tests.json')
        tests_json = json.loads(
            tests_json_string.decode() if six.PY3 else tests_json_string)

        self.assertSequenceEqual(tests_json, test_suite_names)


if __name__ == '__main__':
    unittest.main(verbosity=2)
